// ---------------------------------------------------------------------------
// DeclarationPreprocessor.java
// A class which is dedicated to translate declarations to an intermediate format.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: DeclarationPreprocessor.java,v 1.6 2020/05/25 11:42:22 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import riscal.Main;
import riscal.Main.ErrorWriter;
import riscal.semantics.Types.Value.Int;
import riscal.smt.Utilities.Pair;
import riscal.smt.Utilities.Triple;
import riscal.syntax.AST;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.EnumTypeDefinition;
import riscal.syntax.AST.Declaration.FunctionDeclaration;
import riscal.syntax.AST.Declaration.FunctionDefinition;
import riscal.syntax.AST.Declaration.PredicateDeclaration;
import riscal.syntax.AST.Declaration.PredicateDefinition;
import riscal.syntax.AST.Declaration.PredicateValueDefinition;
import riscal.syntax.AST.Declaration.ProcedureDeclaration;
import riscal.syntax.AST.Declaration.ProcedureDefinition;
import riscal.syntax.AST.Declaration.RecTypeDefinition;
import riscal.syntax.AST.Declaration.TheoremDefinition;
import riscal.syntax.AST.Declaration.TheoremParamDefinition;
import riscal.syntax.AST.Declaration.TypeDefinition;
import riscal.syntax.AST.Declaration.ValueDeclaration;
import riscal.syntax.AST.Declaration.ValueDefinition;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.ImpliesExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.TimesExp;
import riscal.syntax.AST.FunctionSpec;
import riscal.syntax.AST.FunctionSpec.ContractSpec;
import riscal.syntax.AST.FunctionSpec.EnsuresSpec;
import riscal.syntax.AST.FunctionSpec.RequiresSpec;
import riscal.syntax.AST.Identifier;
import riscal.syntax.AST.Multiple;
import riscal.syntax.AST.Parameter;
import riscal.syntax.AST.Specification;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.ASTCloner;
import riscal.tasks.Recursion;
import riscal.tasks.Recursion.CallMap;
import riscal.types.Environment;
import riscal.types.Environment.Symbol;
import riscal.types.Environment.Symbol.FunctionSymbol;
import riscal.types.Environment.Symbol.TypeSymbol;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.types.TypeChecker;

public class DeclarationPreprocessor extends ASTCloner 
{
	/**
	 * A set of strings whose elements shall not be used for new identifier strings
	 */
	private Set<String> identStrings= new HashSet<String>();
	
	/***************************************************************************
	 * Determine whether name is part of identStrings.
	 * @param name The string of interest.
	 * @return true if name is contained in identStrings
	 **************************************************************************/
	boolean forbiddenName(String name)
	{
		return identStrings.contains(name);
	}
	/**
	 * This set collects the strings of all declaration of function-like declarations, 
	 * which have been visited in the current translation
	 */
	private Set<String> visitedIdentStrings=new HashSet<String>();
	
	/**
	 * The names of all types that are introduced in the declarations which shall
	 * be processed.
	 */
	private Set<String> typeIdentStrings=new HashSet<String>();
	
	/**
	 * Assigns to the type symbols new names for the respective type.
	 */
	private Map<TypeSymbol,String> typeRenamings=new HashMap<Environment.Symbol.TypeSymbol, String>();
	
	/***************************************************************************
	 * If typeRenamings indicates that the type, associated to the given symbol, shall be renamed
	 * this method gives the new name, otherwise it returns null
	 * @param sym The symbol associated to the type of interest.
	 * @return The name that shall be used for the type that corresponds to sym, or null.
	 **************************************************************************/
	String getTypeRenaming(TypeSymbol sym)
	{
		String str=typeRenamings.get(sym);
		return str;
	}
	
	/**
	 * Used to collect the names of all variables occuring in the processed declarations.
	 */
	private Set<String> occuringVariables=new HashSet<String>();
	
	/**
	 * Used to identify recursive functions definitions.
	 */
	private CallMap functionCalls;
	
	/***************************************************************************
	 * Determines if the denoted function symbol corresponds to a recursive function.
	 * @param sym The function symbol that corresponds to the function we are interested in.
	 * @return true if the function is recursive.
	 **************************************************************************/
	private boolean isRecursive(FunctionSymbol sym)
	{
		Set<FunctionSymbol> calls=functionCalls.getCalled(sym);
		if(calls==null)
			return false;
		return calls.contains(sym);
	}	
	
	//REMARK: There is no hashfunction nor an equal method for the symbols. Nevertheless we use a hashmap.
	//Since we do not have these methods the default methods defined for java objects are used.
	//This means that only the same instances of the classes are considered equal.
	//But this is OK here since if we have two symbols corresponding to the same entity
	//we have one instance of the object
		
	/**
	 * This Map is used for renaming functions. If we rename a function we assign to the
	 * associated function symbol a string which represents its new name.
	 */
	private Map<FunctionSymbol,String> identRenamings=new HashMap<Environment.Symbol.FunctionSymbol, String>();
	
	/***************************************************************************
	 * If identRenamings indicates that the function, associated to the given symbol, shall be renamed
	 * this method gives the new name, otherwise it returns null.
	 * @param sym The symbol associated to the function of interest.
	 * @return The name that shall be used for the function that corresponds to sym, or null.
	 **************************************************************************/
	String getIdentRenaming(FunctionSymbol sym)
	{
		String str=identRenamings.get(sym);
		return str;
	}
	
	/**
	 * This Map is used for renaming values. If we rename a function we assign to the
	 * associated value symbol a string which represents its new name.
	 */
	private Map<ValueSymbol,String> valueRenamings=new HashMap<Environment.Symbol.ValueSymbol, String>();
	
	/***************************************************************************
	 * If valueRenamings indicates that the value, associated to the given symbol, shall be renamed
	 * this method gives the new name, otherwise it returns null
	 * @param sym The symbol associated to the value of interest.
	 * @return The name that shall be used for the value that corresponds to sym, or null.
	 **************************************************************************/
	String getValueRenaming(ValueSymbol sym)
	{
		String str=valueRenamings.get(sym);
		return str;
	}
	
	/**
	 * This map indicates which functions, represented by their function symbols, were already declared.
	 */	
	private Set<FunctionSymbol> isDeclared=new HashSet<Environment.Symbol.FunctionSymbol>();
	
	/**
	 * Contains the symbols which correspond to the already visited constants.
	 */
	private Set<ValueSymbol> values=new HashSet<Environment.Symbol.ValueSymbol>();
	
	Utilities.Conditions conds;
	
	/**
	 * Contains the names of all functions (excluding functions generated by the system)
	 * that are defined. This includes functions that are introduced by a function definition
	 * and if typeConditionsForFunctionsWithDefiningAxiom is true also functions that
	 * are defined by defining axioms. 
	 */
	private Set<String> definedFunctions=new HashSet<String>();
	
	/**
	 * If declarations shall be ignored this set contains the symbols which correspond to the declarations
	 * which shall not be ignored.
	 */
	private Set<Symbol> keep;
	
	
	/**
	 * A list of strings that shall not be the names of functions/constants/types 
	 * in the preprocessed declarations.
	 */
	private List<String> forbiddenFunctionNames = Arrays.asList(new String[]
			{"BitVec","concat ","extract","bvnot","bvand","bvor","bvneg","bvadd","bvmul","bvudiv","bvurem","bvshl","bvlshr","bvult",
					"bvnand","bvnor","bvxor","bvxnor","bvcomp","bvsub","bvsdiv","bvsrem","bvsmod","bvashr","repeat","zero_extend",
					"sign_extend","rotate_left","rotate_right","bvule","bvugt","bvuge","bvslt","bvsle","bvsgt","bvsge","nat2bv","bv2nat",
					"Bool","true","false","not","and","or","xor","distinct","ite",
					"Array","store","select",
					"assert", "check-sat", "check-sat-assuming", "declare-const",
					"declare-datatype", "declare-datatypes", "declare-fun", "declare-sort", "define-fun",
					"define-fun-rec", "define-sort", "echo", "exit", "get-assertions", "get-assignment", "get-info",
					"get-model", "get-option", "get-proof", "get-unsat-assumptions", "get-unsat-core", "get-value",
					"pop", "push", "reset", "reset-assertions", "set-info", "set-logic", "set-option",
					TypeChecker.resultName
			});

	
	
	/*
	 * If we process a certain declaration we not always get a single new translated declaration but
	 * sometimes we will get besides the actual translated declaration additional ones. Some of these
	 * new Declarations must occur before and some after the actual translated declaration. 
	 * Therefore for convenience I use two different lists for the two kinds of new declarations
	 */
	
	/**
	 * This list contains declarations which were generated in the current translation.
	 * These declarations shall occur before the currently visited declaration.
	 */
	private List<Declaration> generatedDeclarations_Before=new ArrayList<Declaration>();
	
	/**
	 * This list contains declarations which were generated in the current translation.
	 * These declarations shall occur after the currently visited declaration.
	 */
	private List<Declaration> generatedDeclarations_After=new ArrayList<AST.Declaration>();
	


	
	/**
	 * This boolean value indicates whether a theorem shall be processed to
	 * an axiom or to a theorem. If this value is false the theorem is negated.
	 * 
	 * This value will only be false for the theorem that shall be checked.
	 * All other theorems are assumed to hold, thus they can be considered
	 * as axioms. 
	 */
	private boolean isAxiom=true;
	
	boolean processingTheorem()
	{
		return !isAxiom&&!SMT.debug_DoNotNegateGoal;
	}
	
	/**
	 * Indicates if SubTypes were introduced in the processed declarations.
	 */
	private boolean SubTypesPresent=false;
	
	/**
	 * If true, uninterpreted functions are required for the encoding of the
	 * the given RISCAL declarations
	 */
	private boolean uninterpretedFunctionsRequired=false;
	
	/**
	 * If true, non-nullary uninterpreted functions are required for the encoding of the
	 * the given RISCAL declarations
	 */
	private boolean nonNullaryUninterpretedFunctionsRequired=false;

	
	/**
	 * Set uninterpretedFunctionsRequired.
	 * @param b The boolean value that shall be assigned to uninterpretedFunctionsRequired.
	 */
	public void setUninterpretedFunctionsRequired(boolean b)
	{
		uninterpretedFunctionsRequired=b;
	}
	
	/**
	 * Set nonNullaryUninterpretedFunctionsRequired.
	 * @param b The boolean value that shall be assigned to nonNullaryUninterpretedFunctionsRequired.
	 */
	public void setNonNullaryUninterpretedFunctionsRequired(boolean b)
	{
		nonNullaryUninterpretedFunctionsRequired=b;
	}
		
	
	/**
	 * Used to traverse the expressions occurring in the declarations.
	 */
	private ExpressionPreprocessor expPre;
	
	/**
	 * Our standard output stream
	 */
	private static PrintWriter out = Main.getOutput();
	
	/**
	 * Used to type-check the preprocessed declarations.
	 */
	TypeChecker checker=new TypeChecker(new Environment(), new ErrorWriter(out));
	
	/*
	 * The suggestion for the names of defining axioms, specification axioms and theorems.
	 */
	private static final String nameDefiningAxioms="defining_axiom";
	private static final String nameSpecAxioms="spec_axiom";
	private static final String TheoremName="theorem";
	
	/***************************************************************************
	 * The constructed preprocessor may only be applied to declarations that do
	 * not introduces functions/values/types with names not present
	 * in the respective set.
	 * Function calls in declarations have to correspond to funCalls.
	 * Declarations that introduce entities associated to symbols in keep
	 * are ignored.
	 **************************************************************************/
	private DeclarationPreprocessor(Set<String> funcs,Set<String> vals,Set<String> types,CallMap funCalls,Set<Symbol> keep)
	{
		super();
		this.identStrings.addAll(vals);
		this.identStrings.addAll(funcs);
		this.identStrings.addAll(forbiddenFunctionNames);
		this.typeIdentStrings.addAll(types);
		this.visitedIdentStrings.addAll(vals);	
		this.functionCalls=funCalls;
		this.keep=keep;
		if(SMT.considerGuardsForChoices())
			this.conds=new Utilities.Conditions(values);
		expPre=new ExpressionPreprocessor(this,values);
	}
	
	/***************************************************************************
	 * Contains information generated during the preprocessing that
	 * is needed for the translation.
	 **************************************************************************/
	public static class TranslationInformation
	{
		public TranslationInformation(Set<String> names,Set<String> decs,Set<String> defs, boolean ufs,boolean nonNullaryUfs)
		{
			this.forbiddenNames=names;
			this.namesOfDeclarations=decs;
			this.definedFunctions=defs;
			this.uninterpretedFunctionsNecessary=ufs;
			this.nonNullaryUninterpretedFunctionsNecessary=nonNullaryUfs;
		}
		public Set<String> forbiddenNames;
		public Set<String> namesOfDeclarations;
		public Set<String> definedFunctions;
		public boolean uninterpretedFunctionsNecessary;
		public boolean nonNullaryUninterpretedFunctionsNecessary;
	}
	
	/***************************************************************************
	 * Preprocesses a theorem that shall be checked and a list of declarations
	 * it may depend on.
	 * 
	 * Precodnitions:
	 * 	- The declarations must not contain recursive types	
	 * 	- If the declarations contain factorial expressions
	 * 		then the argument of the factorial must be an integer
	 * 		expression with a maximal upper bound of 20.
	 *  - Theorems shall not be used like predicates
	 * @param theorem The theorem that shall be checked
	 * @param declarations The declarations the theorem may depend on.
	 * @return A list containing the preprocessed declarations and
	 * the generated TranslationInformation.
	 **************************************************************************/
	public static Pair<List<Declaration>,TranslationInformation> translate(TheoremParamDefinition theorem, 
			List<Declaration> declarations)
	{
		Triple<Set<String>,Set<String>,Set<String>> idNames=CollectIdentifiers.compute(declarations,theorem);
		//We also have to check for recursions in the theorem -> the theorem could be recursive
		//Note: Theorems should not be used like predicates.
		declarations.add(theorem);
		Specification spec=new Specification(declarations);
		declarations.remove(declarations.size()-1);
		CallMap functionCalls=Recursion.analyze(spec);
		Set<Symbol> s=null;
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
			s=DependingSymbols.giveDependingSymbols(declarations, theorem);
		DeclarationPreprocessor preprocessor=new DeclarationPreprocessor(idNames.first,idNames.second,idNames.third,functionCalls,s);
		List<Declaration> translatedDeclarations=new ArrayList<Declaration>();
		for(Declaration dec:declarations)
		{
			preprocessor.processDec(dec, translatedDeclarations);
		}
		preprocessor.isAxiom=false;
		preprocessor.processDec(theorem, translatedDeclarations);
		translatedDeclarations.addAll(0, preprocessor.generateFactorialDeclarations());
		boolean applySemTrans=preprocessor.SubTypesPresent&&!SMT.expandBaseTypesOfSubTypes;
		riscal.semantics.Translator semTrans=null;
		if(applySemTrans)
		{
			semTrans=new riscal.semantics.Translator(out, true);
		}
		for(Declaration dec:translatedDeclarations)
		{
			preprocessor.checker.process(dec);
			if (preprocessor.checker.hasErrors())
				SMT.illegalUsage(dec, "The type checker reported an error");
			/*
			 * We also have to process functions, as they can occur in the predicates of the subtypes.
			 * We have to ensure that theorems are not translated. As the theorem that we want to show
			 * occurs in negated form. Thus if this theorem is valid the negated form is false
			 * and so the translator will report an error.
			 */
			if(applySemTrans&&!(dec instanceof TheoremDefinition))
			{
				try
				{
					semTrans.process(dec);
				}
				catch(riscal.semantics.Translator.RuntimeError e)
				{
					SMT.illegalUsage(dec, "The semantic translator returned with an error!");
				}
				catch(riscal.semantics.Translator.TranslationError e)
				{
					SMT.illegalUsage(dec, "The semantic translator returned with an error!");
				}
				
			}
		}
		if (preprocessor.checker.hasErrors())
			SMT.illegalUsage(null, "The type checker reported an error");
		preprocessor.identStrings.addAll(preprocessor.occuringVariables);
		TranslationInformation tI=new TranslationInformation(preprocessor.identStrings,
				preprocessor.visitedIdentStrings,preprocessor.definedFunctions,
				preprocessor.uninterpretedFunctionsRequired,preprocessor.nonNullaryUninterpretedFunctionsRequired);
		return new Pair<List<Declaration>,TranslationInformation>(translatedDeclarations,tI);
	}
	
	public static Triple<List<Declaration>, List<List<Declaration>>, TranslationInformation> translateMultiple(
			List<TheoremParamDefinition> theorems,List<Declaration> declarations)
	{
		assert(theorems.size()>0);
		Triple<Set<String>,Set<String>,Set<String>> idNames=CollectIdentifiers.compute(declarations,theorems);
		Specification spec=new Specification(declarations);
		CallMap functionCalls=Recursion.analyze(spec);
		Set<Symbol> s=null;
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			ListIterator<TheoremParamDefinition> it=theorems.listIterator();
			s=DependingSymbols.giveDependingSymbols(declarations, it.next());
			while(it.hasNext())
			{
				s.addAll(DependingSymbols.giveDependingSymbols(declarations, it.next()));
			}		
		}
		DeclarationPreprocessor preprocessor=new DeclarationPreprocessor(idNames.first,idNames.second,idNames.third,functionCalls,s);
		List<Declaration> translatedDeclarations=new ArrayList<Declaration>();
		for(Declaration dec:declarations)
		{
			preprocessor.processDec(dec, translatedDeclarations);
		}
		List<List<Declaration>> theoremDecs=new ArrayList<List<Declaration>>(theorems.size());
		for(TheoremParamDefinition th:theorems)
		{
			preprocessor.isAxiom=false;
			List<Declaration> aux=new ArrayList<AST.Declaration>();
			preprocessor.processDec(th, aux);
			theoremDecs.add(aux);
		}
		translatedDeclarations.addAll(0, preprocessor.generateFactorialDeclarations());
		boolean applySemTrans=preprocessor.SubTypesPresent&&!SMT.expandBaseTypesOfSubTypes;
		riscal.semantics.Translator semTrans=null;
		if(applySemTrans)
		{
			semTrans=new riscal.semantics.Translator(out, true);
		}
		for(Declaration dec:translatedDeclarations)
		{
			preprocessor.checker.process(dec);
			if (preprocessor.checker.hasErrors())
				SMT.illegalUsage(dec, "The type checker reported an error");
			if(applySemTrans&&!(dec instanceof TheoremDefinition))
			{
				try
				{
					semTrans.process(dec);
				}
				catch(riscal.semantics.Translator.RuntimeError e)
				{
					SMT.illegalUsage(dec, "The semantic translator returned with an error!");
				}
				catch(riscal.semantics.Translator.TranslationError e)
				{
					SMT.illegalUsage(dec, "The semantic translator returned with an error!");
				}			
			}
		}
		for(List<Declaration> decs:theoremDecs)
		{
			for(Declaration dec:decs)
			{
				preprocessor.checker.process(dec);
				if (preprocessor.checker.hasErrors())
					SMT.illegalUsage(dec, "The type checker reported an error");
				if(applySemTrans&&!(dec instanceof TheoremDefinition))
				{
					try
					{
						semTrans.process(dec);
					}
					catch(riscal.semantics.Translator.RuntimeError e)
					{
						SMT.illegalUsage(dec, "The semantic translator returned with an error!");
					}
					catch(riscal.semantics.Translator.TranslationError e)
					{
						SMT.illegalUsage(dec, "The semantic translator returned with an error!");
					}			
				}
			}
		}
		if (preprocessor.checker.hasErrors())
			SMT.illegalUsage(null, "The type checker reported an error");
		preprocessor.identStrings.addAll(preprocessor.occuringVariables);
		TranslationInformation tI=new TranslationInformation(preprocessor.identStrings,
				preprocessor.visitedIdentStrings,preprocessor.definedFunctions,
				preprocessor.uninterpretedFunctionsRequired,preprocessor.nonNullaryUninterpretedFunctionsRequired);
		return new Triple<List<Declaration>, List<List<Declaration>>, DeclarationPreprocessor.TranslationInformation>(
				translatedDeclarations, theoremDecs, tI);
			
	}
	
	/***************************************************************************
	 * Preprocesses the given declaration and adds the resulting declaration
	 * plus possible additional declarations to the given list.
	 * @param dec The declaration that shall be preprocessed
	 * @param declst The list to which the results shall be
	 * appended.
	 **************************************************************************/
	private void processDec(Declaration dec,List<Declaration> declst)
	{
		if(SMT.considerGuardsForChoices())
			conds.clearAll();
		Declaration decAux=(Declaration)dec.accept(this);
		declst.addAll(this.generatedDeclarations_Before);
		if(decAux!=null)
		{
			declst.add(decAux);
		}
		declst.addAll(this.generatedDeclarations_After);
		this.generatedDeclarations_Before.clear();
		this.generatedDeclarations_After.clear();
	}
		
	/***************************************************************************
	 * Preprocesses the given expression.
	 * @param exp The expression that shall be preprocessed.
	 * @return The preprocessed expression
	 **************************************************************************/
	private Expression processExpression(Expression exp)
	{
		Expression exp0=expPre.processExpression(exp);
		occuringVariables.addAll(expPre.getVariables());
		generatedDeclarations_Before.addAll(expPre.getGeneratedDeclarations());
		return exp0;
	}
	
	/***************************************************************************
	 * Quantifies the given expression with respect to vars 
	 * (variable, type pairs according to the parameters).
	 * If toProve is
	 * true, then the quantified expression is negated. 
	 * Afterwards this expression gets preprocessed.
	 * @param exp The expression that shall be quantified and preprocessed.
	 * @param vars Contains the variable names/types for the quantification.
	 * @param toProve If true the quantified expression is negated.
	 * @return The preprocessed quantified expression.
	 **************************************************************************/
	private Expression processAndQuantifyExpression(Expression exp,Parameter[] vars,boolean toProve)
	{
		Expression exp0=expPre.processAndQuantifyExpression(exp, vars, toProve);
		occuringVariables.addAll(expPre.getVariables());
		generatedDeclarations_Before.addAll(expPre.getGeneratedDeclarations());
		return exp0;
	}

	
	//--------------------------------------Auxiliaries--------------------------------------//
	
	/***************************************************************************
	 * Generates a function definition for a function representing the factorial.
	 * Shall only be called after the declarations of consideration were
	 * preprocessed.
	 * It is required that only factorials up to 20 occur in the declarations.
	 * Otherwise the result will be erroneous.
	 * @return If the FactorialInformation from the expression preprocessor
	 * indicates that a factorial expression was involved, then
	 * the list contains the necessary declaration for the definition
	 * of a factorial function. Otherwise the list is empty.
	 **************************************************************************/
	private List<Declaration> generateFactorialDeclarations()
	{
		FactorialInformation fI=expPre.getFactorialInfo();
		if(fI.maxIntUpperBound>20)
			SMT.illegalUsage(null, "Only factoials up to 20 are supported!");
		if(fI.factorialOccured)
		{
			String paramName="i";
			Type paramType=new IntType(fI.minIntLowerBound, fI.maxIntUpperBound);			
			paramType.setCanonical();			
			List<Expression> paramExp=new ArrayList<AST.Expression>(1);
			List<Parameter> param0 = new ArrayList<Parameter>(1); 
			paramExp.add(new Expression.MinusExp(new IdentifierExp(new Identifier(paramName)), AST.number(1)));
			param0.add(new Parameter(new Identifier(paramName), paramType));	
			List<Declaration> factorialDeclarations;
			if(SMT.recursiveDefinitions)
			{
				Type factorialType=new IntType(0, factorial(fI.maxIntUpperBound));
				factorialType.setCanonical();
				factorialDeclarations=new ArrayList<AST.Declaration>(1);
				List<FunctionSpec> spec0 = new ArrayList<FunctionSpec>(0); 			
				Expression definingExpression=new TimesExp(new IdentifierExp(new Identifier(paramName)), 
						new Expression.ApplicationExp(new Identifier(fI.name), paramExp));
				definingExpression=new IfThenElseExp(new Expression.EqualsExp(new IdentifierExp(new Identifier(paramName)), 
						AST.number(1)), AST.number(1), definingExpression);
				definingExpression=new IfThenElseExp(new Expression.LessEqualExp(new IdentifierExp(new Identifier(paramName)), 
						AST.number(0)), AST.number(1), definingExpression);
				factorialDeclarations.add(new FunctionDefinition(new Multiple(false), new Identifier(fI.name), param0, 
						factorialType, spec0, definingExpression));
			}
			else
			{				
				Expression definingExpression=AST.number(1);
				long resultTypeLowerBound=1;
				long fac=1;
				if(fI.maxIntUpperBound>1)
				{									
					definingExpression=AST.number(fac);
					for(long i=2;i<=fI.maxIntUpperBound;i++)						
					{
						fac*=i;
						if(i<fI.minIntLowerBound)
						{
							continue;
						}
						else if(i==fI.minIntLowerBound)
						{
							definingExpression=AST.number(fac);
							resultTypeLowerBound=fac;
						}
						else
						{
							definingExpression=new IfThenElseExp(new EqualsExp(new IdentifierExp(new Identifier(paramName)), 
									AST.number(i)), AST.number(fac), definingExpression);
						}					
					}
					if(fI.minIntLowerBound<=1)
					{						
						definingExpression=new IfThenElseExp(new AST.Expression.LessEqualExp(
								new IdentifierExp(new Identifier(paramName)), AST.number(1)),AST.number(1),definingExpression);
					}
				}
				Type factorialType=new IntType(resultTypeLowerBound, fac);
				factorialType.setCanonical();
				List<FunctionSpec> spec0 = new ArrayList<FunctionSpec>(0);
				factorialDeclarations=new ArrayList<AST.Declaration>(1);
				factorialDeclarations.add(new FunctionDefinition(new Multiple(false), new Identifier(fI.name), param0, 
						factorialType, spec0, definingExpression));
			}
			return factorialDeclarations;
		}
		else
		{
			return new ArrayList<AST.Declaration>(0);
		}
	}
		
	/***************************************************************************
	 * Computes the factorial
	 * @param x x<=20
	 * @return The factorial of x
	 **************************************************************************/
	private static long factorial(long x)
	{
		if(x<=1)
			return 1;
		else
			return x*factorial(x-1);	
	}
	
	//----------------------------Special handling for certain entities----------------------------//
	
	/***************************************************************************
	 * Computes an identifier whose name is a string that is not part of identStrings.
	 * This string is added to identStrings.
	 * @param suggestion A basis for the string that shall be constructed.
	 * @return The described identifier.
	 **************************************************************************/
	private Identifier generateNewIdent(String suggestion)
	{
		return new Identifier(Utilities.NewIdentString(suggestion,identStrings));
	}
	
	/***************************************************************************
	 * Computes a string that is not part of identStrings.
	 * The generated string is added to identStrings.
	 * @param suggestion A basis for the string that shall be constructed.
	 * @return The given string, appended with a suitable number such
	 * that the resulting string is not part of identStrings
	 **************************************************************************/
	String giveNewName(String suggestion)
	{
		return Utilities.NewIdentString(suggestion,identStrings);
	}
	
	/***************************************************************************
	 * Computes a string that is not part of identStrings.
	 * The generated string is added to identStrings.
	 * Add the string to definedFunctions.
	 * @param suggestion A basis for the string that shall be constructed.
	 * @return The given string, appended with a suitable number such
	 * that the resulting string is not part of identStrings
	 **************************************************************************/
	String giveFunctionNameAndSetAsDefined(String suggestion)
	{
		String name=giveNewName(suggestion);
		definedFunctions.add(name);
		return name;
	}
	
	/***************************************************************************
	 * Computes a string that is neither part of identStrings nor of constraints.
	 * The generated string is added to constraints.
	 * @param suggestion A basis for the string that shall be constructed.
	 * @return The given string, appended with a suitable number such
	 * that the resulting string is not part of identStrings or constraints.
	 **************************************************************************/
	String giveNewNameConstraint(String suggestion,Set<String> constraints)
	{
		return Utilities.NewIdentStringWithConstraints(suggestion,constraints,identStrings);
	}
	

	/***************************************************************************
	 * The method extracts the pre -and postconditions part of the given
	 * array of specification. 
	 * The method does not clone the expression! The caller is responsible for cloning!
	 * @param funspec The function specification that shall be processed.
	 * @return A conjunction of the preconditions and a conjunction of the
	 * postconditions
	 **************************************************************************/
	private Pair<Expression,Expression> collectExpressionsFromSepecifications(FunctionSpec[] funspec)
	{
		Expression requires=null;
		Expression ensures=null;
		for(FunctionSpec s:funspec)
		{
			if(s instanceof FunctionSpec.RequiresSpec)
			{
				requires=AST.and(requires, processRequiresSpec((RequiresSpec) s));
			}
			else if(s instanceof FunctionSpec.EnsuresSpec)
			{
				ensures=AST.and(ensures, processEnsuresSpec((EnsuresSpec) s));
			}
		}
		return new Pair<Expression,Expression>(requires,ensures);
	}
	
	private Pair<Expression,Expression> giveSpec(FunctionSpec[] funspec)
	{
		return collectExpressionsFromSepecifications(funspec);
	}
	
	private void analyseFunSpecForConditions(FunctionSpec[] funspec)
	{
		if(SMT.considerGuardsForChoices())
		{
			for(FunctionSpec sp:funspec)
			{
				if(sp instanceof RequiresSpec)
				{
					Expression condition=((RequiresSpec)sp).exp;
					conds.analyseWithSimpleConditions(condition);
				}
			}
		}		
	}
	
	
	/***************************************************************************
	 * The method extracts the expression of a requires specification
	 **************************************************************************/
	private Expression processRequiresSpec(FunctionSpec.RequiresSpec spec)
	{
		return spec.exp;	
	}
	
	/***************************************************************************
	 * The method extracts the expression of an ensures specification
	 **************************************************************************/
	private Expression processEnsuresSpec(FunctionSpec.EnsuresSpec spec)
	{
		return spec.exp;
	}
	
	/***************************************************************************
	 * Determines whether the given specifications contain a ContractSpec.
	 * @param funspec The specifications that shall be analysed.
	 * @return true if the given specifications contain a ContractSpec.
	 **************************************************************************/
	private static boolean hasContractSpec(FunctionSpec[] funspec)
	{
		return Arrays.stream(funspec).anyMatch(spec -> spec instanceof ContractSpec);
	}
	
	/***************************************************************************
	 * Determines if a function with the given functionSpec shall
	 * be treated in a modular way.
	 * @param funspec The FunctionSpec of the function we are
	 * interested in.
	 * @return true if SMT.functionDefinitionTreatmentMethod=Modular
	 * 	or SMT.functionDefinitionTreatmentMethod=ModularAccordingToSpecification
	 * 		and the given FunctionSpec contains a ContractSpec
	 **************************************************************************/
	static boolean applyModularReasoning(FunctionSpec[] funspec)
	{
		switch(SMT.functionDefinitionTreatmentMethod)
		{
		case PreserveDefinitions:
			return false;
		case Modular:
			return true;
		case ModularAccordingToSpecification:
			return hasContractSpec(funspec);
		}
		return false;
	}	

	/***************************************************************************
	 * This method is used to preprocess a function or a predicate definition.
	 * Can also be applied to TheoremParamDefinition for the case that it should be
	 * possible to call theorems like predicates (But this should not happen)
	 * @param sym The symbol of the defined function/predicate.
	 * @param returnType The result type of the defined function.
	 * @param params The parameters of the defined function.
	 * @param treatModular If true: introduce the function via its specification
	 * @param funspec The specification of the defined function.
	 * @param definingExp The defining expression of the function.
	 * @param mult Multiple of the definition.
	 * @return Preprocessed definition.
	 **************************************************************************/
	private Declaration processFunctionDef(FunctionSymbol sym,Type returnType,Parameter[] params,boolean treatModular,
			FunctionSpec[]funspec,Expression definingExp,Multiple mult){
		String functionName=sym.ident.string;
		Identifier ident0;
		if(visitedIdentStrings.contains(functionName)||forbiddenFunctionNames.contains(functionName))
		{
			if(isDeclared.contains(sym))//in this case the function was already declared
			{
				setUninterpretedFunctionsRequired(true);
				if(params.length>0)
					setNonNullaryUninterpretedFunctionsRequired(true);
				String nameToUse=getIdentRenaming(sym);
				nameToUse=nameToUse==null?sym.ident.string:nameToUse;
				ident0=new Identifier(nameToUse);
				if(treatModular)
				{
					return giveSpecifingAxiom(ident0,params, funspec);
				}
				else
				{
					return giveDefiningAxiom(ident0, params, definingExp);
				}
			}
			//the function was not declared, therefore another function with the same name already occurred -> renaming
			else
			{
				String newName=Utilities.NewIdentString(functionName, this.identStrings);
				identRenamings.put(sym, newName);
				ident0=new Identifier(newName);
			}
		}
		else
		{
			ident0=(Identifier) sym.ident.accept(this);
		}
		visitedIdentStrings.add(ident0.string);
		if(treatModular)
		{
			setUninterpretedFunctionsRequired(true);
			if(params.length>0)
				setNonNullaryUninterpretedFunctionsRequired(true);
			Type type0=expPre.translateType(returnType);
			List<Parameter> params0=Arrays.stream(params).map(p -> (Parameter)p.accept(this)).collect(Collectors.toList());
			Declaration funDec=new FunctionDeclaration(mult, ident0, params0, type0);	
			generatedDeclarations_Before.add(funDec);
			return giveSpecifingAxiom(ident0,params, funspec);
		}
		else
		{
			if(isRecursive(sym))
			{
				setUninterpretedFunctionsRequired(true);
				if(params.length>0)
					setNonNullaryUninterpretedFunctionsRequired(true);
				Type type0=expPre.translateType(returnType);
				List<Parameter> params0=Arrays.stream(params).map(p -> (Parameter)p.accept(this)).collect(Collectors.toList());
				Declaration funDec=new FunctionDeclaration(mult, ident0, params0, type0);	
				generatedDeclarations_Before.add(funDec);
				return giveDefiningAxiom(ident0, params, definingExp);
			}
			else	
			{
				Type type0=expPre.translateType(returnType);
				return giveFunctionDefinition(ident0,type0,params,definingExp,mult);
			}
		}
	}
	
	/***************************************************************************
	 * Applies a function given by ident to variables given by the identifiers of the parameters.
	 * @param ident A identifier corresponding to a function.
	 * @param params Parameters whose types correspond to the arity of the function given by ident.
	 * @return The described function application.
	 **************************************************************************/
	private Expression giveApplicationExp(Identifier ident,Parameter[] params)
	{
		List<Expression> exps=Arrays.stream(params).map(p -> new IdentifierExp(p.ident)).collect(Collectors.toList());		
		return new ApplicationExp(ident, exps);
	}
	
	/***************************************************************************
	 * Generates the defining axiom for a given function.
	 * 		Example: 					fun f(x:Nat[4]):Nat[4]=x
	 * 		defining theorem:			forall x:Nat[4]. f(x)=x
	 * @param ident the identifier corresponding to the function that
	 * shall be defined.
	 * @param params The parameters of the function.
	 * @param definingExp The defining expression of the function.
	 * @return A TheoremDefinition representing the defining axiom.
	 **************************************************************************/
	private Declaration giveDefiningAxiom(Identifier ident,Parameter[] params,Expression definingExp)
	{
		if(!SMT.typeConditionsForFunctionsWithDefiningAxiom)
			definedFunctions.add(ident.string);
		Expression exp0=giveApplicationExp(ident, params);
		exp0=new EqualsExp(exp0, definingExp);
		exp0=processAndQuantifyExpression(exp0, params, false);
		return new TheoremDefinition(generateNewIdent(ident.string+"_"+nameDefiningAxioms), exp0, true);
	}
	
	/***************************************************************************
	 * Generates a FunctionDefinition
	 * @param ident The identifier of the function that shall be defined.
	 * @param returnType The type of the function.
	 * @param params The parameters of the function.
	 * @param definingExp The defining expression of the function,
	 * @param mult Mult of the function.
	 * @return A FunctionDefinition incorporating the
	 * information from the arguments of this method.
	 **************************************************************************/
	private Declaration giveFunctionDefinition(Identifier ident,Type returnType,Parameter[] params,Expression definingExp,Multiple mult)
	{
		definedFunctions.add(ident.string);
		List<Parameter> params0=Arrays.stream(params).map(p -> (Parameter)p.accept(this)).collect(Collectors.toList());
		Type type0=(Type) returnType.accept(this);
		List<FunctionSpec> spec0=new ArrayList<FunctionSpec>(0);
		Expression exp0=processExpression(definingExp);
		return new FunctionDefinition(mult, ident, params0, type0, spec0, exp0);
	}
	
	/***************************************************************************
	 * Generates the specifying axiom for a function.
	 * Example 	fun f(x:Nat[4]):Nat[4]
	 * 					requires x>0;
	 * 					ensures result<1;
	 * -> forall x:Nat[4]. (x>0 => f(x)<1)
	 * @param ident the identifier corresponding to the function that
	 * shall be defined.
	 * @param params The parameters of the function.
	 * @param funspec The specification of the function.
	 * @return A TheoremDefinition representing the specifying axiom.
	 **************************************************************************/
	private Declaration giveSpecifingAxiom(Identifier ident,Parameter[] params,FunctionSpec funspec[])
	{
		Pair<Expression,Expression> spec=giveSpec(funspec);
		if(spec.second==null)
			return null;
		Expression exp0;
		
		Expression post=spec.second;
		Set<ValueSymbol> free=TypedFreeVariables.compute(spec.second);
		for(ValueSymbol sym:free)
		{
			if(sym.ident.string.equals(TypeChecker.resultName))
			{
				Identifier letId=new Identifier(TypeChecker.resultName);
				//the identifier needs the appropiate symbol in order to make the renaming possible
				letId.setSymbol(sym);
				Binder resultBinder=new Binder(letId,giveApplicationExp(ident, params));
				List<Binder> lst=new ArrayList<AST.Binder>(1);
				lst.add(resultBinder);
				post=new LetParExp(lst, post);
			}
		}	
		if(spec.first==null)
		{
			exp0=processAndQuantifyExpression(post, params, false);
		}
		else
		{
			exp0=processAndQuantifyExpression(new riscal.syntax.AST.Expression.ImpliesExp(spec.first,post), params, false);
		}
		return new TheoremDefinition(generateNewIdent(ident.string+"_"+nameSpecAxioms), exp0, true);
	}
	
	/***************************************************************************
	 * This method is used to preprocess a function/predicate/procedure declaration.
	 * @param ident The identifier of the declared entity.
	 * @param resultType The type of the declared entity.
	 * @param params The parameters of the declared entity.
	 * @param mult
	 * @return A FunctionDeclaration representing the preprocessed entity.
	 **************************************************************************/
	private Declaration processFunctionDeclaration(Identifier ident,Type resultType,Parameter[] params,Multiple mult)
	{
		setUninterpretedFunctionsRequired(true);
		if(params.length>0)
			setNonNullaryUninterpretedFunctionsRequired(true);
		String funName=ident.string;
		List<Parameter> param0 = new ArrayList<Parameter>(params.length);
		for (Parameter p : params) param0.add((Parameter)p.accept(this));
		Type type0 = expPre.translateType(resultType);
		if(visitedIdentStrings.contains(funName)||forbiddenFunctionNames.contains(funName))
		{
			Identifier newIdent=generateNewIdent(funName);
			visitedIdentStrings.add(newIdent.string);
			identRenamings.put((FunctionSymbol) ident.getSymbol(), newIdent.string);
			return new FunctionDeclaration(mult, newIdent, param0, type0);
		}
		else//the function name was not already used
		{
			isDeclared.add((FunctionSymbol) ident.getSymbol());
			visitedIdentStrings.add(funName);
			return new FunctionDeclaration(mult, new Identifier(funName), param0, type0);			
		}
	}
	
	/***************************************************************************
	 * Contains the necessary information to introduce a
	 * function representing the factorial.
	 **************************************************************************/
	public static class FactorialInformation
	{
		public boolean factorialOccured;
		public long maxIntUpperBound;
		public long minIntLowerBound;
		public String name;
		public FactorialInformation(String name)
		{
			factorialOccured=false;
			maxIntUpperBound=1;
			minIntLowerBound=1;
			this.name=name;
		}
	}
	
	
	//--------------------------------------Visit Methods--------------------------------------//
	
	//*************************************Declarations****************************************//
	
	public Declaration visit(ValueDeclaration decl)
	{
		values.add((ValueSymbol)decl.ident.getSymbol());
		int val=((Int)((ValueSymbol)(((ValueDeclaration) decl).ident).getSymbol()).getValue()).getValue();
		Expression exp0=AST.number(val);
		Identifier ident0;
		if(forbiddenFunctionNames.contains(decl.ident.string))
		{
			String valName=Utilities.NewIdentString(decl.ident.string, this.identStrings);
			ident0=new Identifier(valName);
			valueRenamings.put((ValueSymbol) decl.ident.getSymbol(), valName);
		}
		else
		{
			ident0=(Identifier) decl.ident.accept(this);
		}
		Type type0=new Type.IntType(AST.number(val),AST.number(val));
		return new ValueDefinition(ident0, type0, exp0);	
	}
	
	public Declaration visit(FunctionDeclaration decl)
	{
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Type t=((FunctionSymbol)decl.ident.getSymbol()).type;
		return processFunctionDeclaration(decl.ident, t, decl.param, decl.multiple);
//		return processFunctionDeclaration(decl.ident, decl.type, decl.param, decl.multiple);
	}
	
	public Declaration visit(PredicateDeclaration decl)
	{
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		return processFunctionDeclaration(decl.ident, new BoolType(), decl.param, decl.multiple);
	}
	
	public Declaration visit(ProcedureDeclaration decl)
	{
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Type t=((FunctionSymbol)decl.ident.getSymbol()).type;
		return processFunctionDeclaration(decl.ident, t, decl.param, decl.multiple);
//		return processFunctionDeclaration(decl.ident, decl.type, decl.param, decl.multiple);
	}
	
  public Declaration visit(TypeDefinition decl)
  {
  	Identifier ident0;
  	//boolector does not allow sorts with names like "store" or "and"
  	//Therefore we have to rename such types.
  	if(forbiddenFunctionNames.contains(decl.ident.string))
  	{
  		String newName=Utilities.NewIdentString(decl.ident.string, typeIdentStrings);
  		typeRenamings.put((TypeSymbol) decl.ident.getSymbol(), newName);
  		ident0=new Identifier(newName);
  	}
  	else
  	{
  		ident0 = (Identifier)decl.ident.accept(this);
  	}
  	if(((TypeSymbol)decl.ident.getSymbol()).getType() instanceof SubType)
  	{
  		SubTypesPresent=true;
  	}
//  	Type type0=expPre.translateType(((TypeSymbol)decl.ident.getSymbol()).getType());
  	Type t=((TypeSymbol)decl.ident.getSymbol()).getType();
  	Type type0;
  	if(t instanceof SubType)
  	{
  		type0 = expPre.translateType(decl.type);
  	}
  	else
  	{
  		//If we would treat this case like the above one we would get a problem
  		//with declarations like "type t=Nat[N!];"
  		type0=expPre.translateType(t);
  	} 	
    Expression exp0 = decl.exp == null ? null : processExpression(decl.exp);
    return new TypeDefinition(ident0, type0, exp0);
  }
	
	public Declaration visit(RecTypeDefinition decl)
	{
		SMT.notImplemented(decl, "Recursive Types are not yet implemented");
		return null;
	}
	
	public Declaration visit(EnumTypeDefinition decl)
	{
		SMT.notImplemented(decl, "Enum Types are not yet implemented");
		return null;
	}

	public Declaration visit(ValueDefinition decl)
	{
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Identifier ident0;
		if(forbiddenFunctionNames.contains(decl.ident.string))
		{
			String valName=Utilities.NewIdentString(decl.ident.string, this.identStrings);
			ident0=new Identifier(valName);
			valueRenamings.put((ValueSymbol) decl.ident.getSymbol(), valName);
		}
		else
		{
			ident0=(Identifier) decl.ident.accept(this);
		}
		values.add((ValueSymbol)decl.ident.getSymbol());
		Expression exp0=processExpression(decl.exp);
		Type type0 = decl.type == null ? null : expPre.translateType(((ValueSymbol)decl.ident.getSymbol()).type);
		return new ValueDefinition(ident0, type0, exp0);
	}
	
	public Declaration visit(PredicateValueDefinition decl)
	{
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Identifier ident0;
		if(forbiddenFunctionNames.contains(decl.ident.string))
		{
			String valName=Utilities.NewIdentString(decl.ident.string, this.identStrings);
			ident0=new Identifier(valName);
			valueRenamings.put((ValueSymbol) decl.ident.getSymbol(), valName);
		}
		else
		{
			ident0=(Identifier) decl.ident.accept(this);
		}
		values.add((ValueSymbol)decl.ident.getSymbol());
		Expression exp0=processExpression(decl.exp);
		return new ValueDefinition(ident0, new Type.BoolType(), exp0);
	}		
	
	public Declaration visit(TheoremDefinition decl)
	{
		if(isAxiom&&(SMT.skipDeclarations!=SMT.SkipMode.RetainAll))
		{			
			if(SMT.skipDeclarations==SMT.SkipMode.SkipUnused)
				return null;
			if(!decl.axiom&&SMT.skipDeclarations==SMT.SkipMode.RetainAxioms)
				return null;
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Expression exp0=processExpression(decl.exp);
		Identifier ident0;
		if(forbiddenFunctionNames.contains(decl.ident.string))
		{
			String valName=Utilities.NewIdentString(decl.ident.string, this.identStrings);
			ident0=new Identifier(valName);
			valueRenamings.put((ValueSymbol) decl.ident.getSymbol(), valName);
		}
		else
		{
			ident0=(Identifier) decl.ident.accept(this);
		}
		if(SMT.treatTheoremsAsFunctions)
		{
			String valDefName=Utilities.NewIdentString(ident0.string+TheoremName, identStrings);
			Identifier valIdent=new Identifier(valDefName);
			Declaration valueDec=new ValueDefinition(valIdent, new Type.BoolType(), exp0);
			generatedDeclarations_Before.add(valueDec);
			values.add((ValueSymbol)decl.ident.getSymbol());
			return new TheoremDefinition(ident0, new IdentifierExp(valIdent), isAxiom);
		}
		return new TheoremDefinition(ident0, exp0, isAxiom);		
	}
	
	public Declaration visit(FunctionDefinition decl)
	{		
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Type t=((FunctionSymbol)decl.ident.getSymbol()).type;
		return processFunctionDef((FunctionSymbol) decl.ident.getSymbol(),t,decl.param,
				applyModularReasoning(decl.spec),decl.spec,decl.exp,decl.multiple);
	}
	
  public Declaration visit(PredicateDefinition decl)
  {
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		return processFunctionDef((FunctionSymbol) decl.ident.getSymbol(),new BoolType(),decl.param,
				applyModularReasoning(decl.spec),decl.spec,decl.exp,decl.multiple);
  }
  
  public Declaration visit(ProcedureDefinition decl)
  {
		if(SMT.skipDeclarations!=SMT.SkipMode.RetainAll)
		{
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		Type t=((FunctionSymbol)decl.ident.getSymbol()).type;
		return processFunctionDef((FunctionSymbol) decl.ident.getSymbol(),t,decl.param,
				true,decl.spec,null,decl.multiple);
  }
  
	public Declaration visit(TheoremParamDefinition decl)
	{		
//		if(SMT.ignoreAxioms&&isAxiom)
//			return null;
		if(isAxiom&&(SMT.skipDeclarations!=SMT.SkipMode.RetainAll))
		{
			if(SMT.skipDeclarations==SMT.SkipMode.SkipUnused)
				return null;
			if(!decl.axiom&&SMT.skipDeclarations==SMT.SkipMode.RetainAxioms)
				return null;
			if(!keep.contains(decl.ident.getSymbol()))
				return null;
		}
		if(SMT.treatTheoremsAsFunctions)
		{
			Declaration x=processFunctionDef((FunctionSymbol) decl.ident.getSymbol(),new BoolType(),decl.param,
					applyModularReasoning(decl.spec),decl.spec,decl.exp,decl.multiple);
			generatedDeclarations_Before.add(x);
			Identifier id=new Identifier(Utilities.NewIdentString(decl.ident.string+"_"+TheoremName, identStrings));	
			String name=identRenamings.get(decl.ident.getSymbol());
			if(name==null)
			{
				name=decl.ident.string;
			}
			Identifier theoremFunctionIdent=new Identifier(name);
			Expression exp0=giveApplicationExp(theoremFunctionIdent, decl.param);
			Expression preCond=giveSpec(decl.spec).first;
			if(preCond!=null)
			{
				exp0=new ImpliesExp(preCond, exp0);
			}
			exp0=processAndQuantifyExpression(exp0, decl.param, !isAxiom);
			return new TheoremDefinition(id, exp0, isAxiom);
		}
		analyseFunSpecForConditions(decl.spec);
		String name=decl.ident.string;
		if(visitedIdentStrings.contains(name)||forbiddenFunctionNames.contains(name))
		{
			name=Utilities.NewIdentString(name, identStrings);
		}		
		visitedIdentStrings.add(name);
		Identifier id=new Identifier(name);	
		Expression preCond=giveSpec(decl.spec).first;
		Expression definingexp=decl.exp;
		if(preCond!=null)
		{
			definingexp=new ImpliesExp(preCond, definingexp);
		}
		definingexp=processAndQuantifyExpression(definingexp, decl.param, !isAxiom);
		return new TheoremDefinition(id, definingexp, isAxiom);
	}
	
  public Parameter visit(Parameter param)
  {
    Identifier ident0 = (Identifier)param.ident.accept(this);
    Type type0 = expPre.translateType(((ValueSymbol)param.ident.getSymbol()).type);
//    Type type0 = translateType(param.type);
    return new Parameter(ident0, type0); 
  }
	
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
