// ---------------------------------------------------------------------------
// SMT.java
// class for the translation to SMT-LIB scripts.
// $Id: SMT.java,v 1.18 2020/06/24 09:54:53 schreine Exp $
//
// Authors: Franz-Xaver Reichl <franz.x.reichl@gmail.com>
//      and Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.smt;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Consumer;

import riscal.Main;
import riscal.smt.DeclarationPreprocessor.TranslationInformation;
import riscal.smt.Utilities.Pair;
import riscal.smt.Utilities.Triple;
import riscal.syntax.AST;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.TheoremParamDefinition;
import riscal.syntax.AST.Type;

/*
 * Reminder
 * - This translation relies on the order of the value generation as of RISCAL 2.10.5
 * 
 * - Applying the translation several times to the same input may generate different
 * 			SMT-LIB scripts. The order of quantified variables in the generated axioms
 * 			for choose/min/max expressions may have a different order.
 */

public class SMT
{
	/***************************************************************************
	 * If true, calling the SMT solver in the task menu is enabled
	 **************************************************************************/
  public static boolean enabled = true;
  
  public static boolean enableiterativeTheoremChecking=true;
  
  /***************************************************************************
	 * If true, the SMT menu provides debug options.
	 **************************************************************************/
  public static final boolean enableDebugOptions=false;
  
  /***************************************************************************
	 * If true, the SMT menu provides experimental options.
	 * This includes an Option for the usage of SMT-LIB quantifiers
	 **************************************************************************/
  public static final boolean enableExperimentalOptions=true;
  
  /***************************************************************************
   * If true, and SMT-LIB quantifiers shall be used also uninterpreted functions
   * can be used. (Only if the selected solvers supports UFBV)
   **************************************************************************/
  private static boolean SMT_LIB_Quantifiers_UseUFs=true;
  
  /***************************************************************************
   * Experimental
   * If true, the SMT solver is used to check whether the dependencies
   * of the theorem that shall be checked are satisfiable. If the are
   * unsatisfiable either a wrong option was used, a proof obligation
   * that should be valid is invalid or an error in the implementation
   * of the translation was found
   **************************************************************************/
  public static boolean checkDependencies=false;
   
  public enum FunctionHandlingMode {PreserveDefinitions,Modular,ModularAccordingToSpecification};
  private static final FunctionHandlingMode defaultFunctionDefinitionTreatmentMethod=FunctionHandlingMode.PreserveDefinitions;
  
  /***************************************************************************
   * Determines how to handle non recursive function and predicate definitions
   * If this variable is set to
   * - PreserveDefinitions
   *			The definitions of functions(predicates) are preserved, i.e. depending on the situation either
   * 			defining axioms are generated or the definitions are kept.
   *  - Modular
   *			Axioms are generated from the specification of a function
   *  - ModularAccordingToSpecification
   *  	  If the specification of a function(predicate) contains a contractSpec an axiom is generated from the spec
   *  		otherwise the definition is preserved
   **************************************************************************/
  public static FunctionHandlingMode functionDefinitionTreatmentMethod=defaultFunctionDefinitionTreatmentMethod;
  
  /***************************************************************************
   * If true, recursive functions are translated to define-fun-rec 
   * expressions.
   * Note: This option should be set to false since:
   * 		Not all the supported SMT solvers can handle the keyword define-fun-rec
   * 		Because of the above reason this option was not tested.
   **************************************************************************/
  public static boolean recursiveDefinitions = false;
  
  /***************************************************************************
   * If true, universal and existential quantifiers get removed.
   * Note: This option should be set to true since:
   * 		Not all the supported SMT solvers can handle the quantifiers and uninterpreted functions.
   * 		Because of the above reason this option was not sufficiently tested.
   **************************************************************************/
  private static boolean unquantify=true;
  
  /***************************************************************************
   * @return true if quantifiers shall be removed.
   **************************************************************************/
  public static boolean getUnquantify()
  {
  	return unquantify;
  }
  
  /***************************************************************************
   * @param b true if universal and existential quantifier shall be removed.
   **************************************************************************/
  public static void setUnquantify(boolean b)
  {
  	unquantify=b;
//  	if(!b)
//  		transformToNegationNormalForm=false;
//  	else
//  	{
//  		if(skolemMode==SkolemisationMode.NeverSkolemise)
//  			transformToNegationNormalForm=false;
//  		else
//  			transformToNegationNormalForm=true;
//  	}
  }
  
  
  public enum SkolemisationMode {AlwaysSkolemise,SkolemiseUseHeuristic,NeverSkolemise};
  private static final SkolemisationMode defaultSkolemMode=SkolemisationMode.SkolemiseUseHeuristic;
  
  /***************************************************************************
   * Determines if skolemisation shall be applied to eligible existential quantifiers.
   * If this variable is set to
   *  - AlwaysSkolemise
   *  		Apply skolemisation to every eligible existantial quantifier
   *  - SkolemiseUseHeuristic
   *  		Apply skolemisation to eligible existential quantifier with respect to a heuristic.
   *  		Expand the other existential quantifiers.
   *  - NeverSkolemise
   *  		Expand every existential quantifier
   **************************************************************************/
  private static SkolemisationMode skolemMode=defaultSkolemMode;
  
//  /***************************************************************************
//   * If unquantify and this variable are true existential quantifiers are 
//   * skolemised instead and not expanded (if possible)
//   **************************************************************************/
//  private static boolean useSkolemisation=true;
  
  /***************************************************************************
   * @return true if skolemisation shall be applied.
   **************************************************************************/
  public static SkolemisationMode getUseSkolemisation()
  {
  	return skolemMode;
  }
  
  /***************************************************************************
   * Sets the skolemisation option.
   * @param b true if skolemisation shall be applied.
   **************************************************************************/
  public static void setUseSkolemisation(SkolemisationMode s)
  {
  	skolemMode=s;
  	if(s==SkolemisationMode.NeverSkolemise)
  		transformToNegationNormalForm=false;
  	else
  		transformToNegationNormalForm=true;
  }
  
  /***************************************************************************
   * If true formulae are transformed to negation normal form.
   * Note: should not be set to false. Skolemisation requires negation normal forms.
   **************************************************************************/
  private static boolean transformToNegationNormalForm=true;
  
  /***************************************************************************
   * @return true if formulae shall be transfered to negation normal form.
   **************************************************************************/
  public static boolean getTransformToNegationNormalForm()
  {
  	return transformToNegationNormalForm;
  }
  
  /***************************************************************************
   * If true equivalences do not get transformed to negation normal form.
   * Idea behind this option:
   * If there is an quantifier in an equivalence we have to expand it anyways.
   * So maybe it is better not to rewrite equivalences.
   **************************************************************************/
  public static boolean skolemisationInEquiavalences=false;
  
  /***************************************************************************
   * If true ParameterTheorems are considered as predicates which are true for all applicable parameters
   * and Theorems are considered as Boolean Values which are true for all applicable parameters.
   * Note should be false. The option true was not sufficiently tested.
   **************************************************************************/
  public static boolean treatTheoremsAsFunctions=false;
  
  /***************************************************************************
   * If this is true, all the values of the base type are generated instead of just the values of the subtype.
   * The reason for this option is:
   * For expanding subtypes we need the semantic predicate of the subtype. But the question is if we really should
   * generate the semantic entities.
   * Note:
   * This option is depreciated, set it to false. 
   * If this value is set to true at least in combination with sets errors may occur.
   **************************************************************************/
  public static boolean expandBaseTypesOfSubTypes=false;
  
  /***************************************************************************
   * If true the tuple access respectively the tuple update functions for each tuple type
   * are translated to a single operation. Otherwise there is for each index a separate function.
   * Note should be set to false.
   **************************************************************************/
  public static boolean treatTupleOperationsAsSingleOperations=false;
  
  /***************************************************************************
   * If true the enumeration function for tuples and maps is an uninterpreted
   * function, that is defined by asserts.
   **************************************************************************/
  public static boolean defineEnumerationForTuplesAndMapsByTables=false;
  
  /***************************************************************************
   * If true then for function like entities 
   * that are defined via a defining axiom
   * a condition is introduced which ensures that its translation
   * has the appropriate type.
   **************************************************************************/
  public static boolean typeConditionsForFunctionsWithDefiningAxiom=false;
  
  /***************************************************************************
   * If true the different values of the type are directly plugged into
   * the function. Otherwise "lets" are used similar to the expansion
   * of quantifiers without auxiliary functions.
   **************************************************************************/
  public static boolean plugValuesDirectlyIntoAuxiliaryFunctions=true;
  
  public enum UsageModeGuardsForChoices {NoGuards,GuardsForSimpleConditions,UseGuards};
  public static UsageModeGuardsForChoices guardChoices=UsageModeGuardsForChoices.NoGuards;
	public static boolean considerGuardsForChoices()
	{
		if(SMT.guardChoices==SMT.UsageModeGuardsForChoices.NoGuards)
			return false;
		else
			return true;
	}
	
	/***************************************************************************
	 * If true uninterpreted functions are used for the representation of
	 * min and max expressions similarly as for choose expressions.
   **************************************************************************/
	public static boolean representMinMaxByUninterpretedFunctions=false;
  
  public enum QuantifierOrdering {NoOrdering,OrderAscending_TypeSize,OrderDescending_TypeSize};
  private static final QuantifierOrdering defaultQvOrdering=QuantifierOrdering.OrderDescending_TypeSize;
  
  /***************************************************************************
   * Determines the order of variables when a quantified variable
   * with multiple variables is unnested.
   * 
   * - NoOrdering: An arbitrary order of the variables.
   * 		Note: This especially means that if a specification contains
   * 		for example choose expressions the resulting SMT-LIB
   * 		script can differ from run to run.
   * - OrderAscending_TypeSize: The first variable is the variable
   * 		with the smallest type size. The remaining variables
   * 		are ordered ascending with respect to the type size.
   * - OrderDescending_TypeSize: The first variable is the variable
   * 		with the largest type size. The remaining variables
   * 		are ordered descending with respect to the type size.
   **************************************************************************/
  public static QuantifierOrdering qvOrdering=defaultQvOrdering;
  
  /***************************************************************************
   * Determines the maximal size of a type which can be treated.
   **************************************************************************/
  public static int maximalTypeSize=Integer.MAX_VALUE;   
  
  /***************************************************************************
   * If true the output of the SMT solver and some timings are printed to the PrintWriter out.
   **************************************************************************/
  private static boolean printResult=true;
  /***************************************************************************
   * If true the intermediate result of the preprocessor is printed.
   **************************************************************************/
  public static boolean diagnosticPrints=false;
  /***************************************************************************
   * If true the resulting SMT-LIB script is written to a file with a name
   * determined by fileNameSMT_LIBScript in the current directory.
   **************************************************************************/
  private static final boolean defaultWriteTranslatedSpecToFile=false;
  public static boolean writeTranslatedSpecToFile=defaultWriteTranslatedSpecToFile;
  
  /***************************************************************************
   * If writeTranslatedSpecToFile is true then the resulting SMT-LIB
   * script is written to a file with the name fileNameSMT_LIBScript
   * in the directory dirSMT_LIBScript.
   * If changed the caller is responsible to ensure that the new name
   * is a valid filename.
   **************************************************************************/
  private static final String defaultFileNameSMT_LIBScript="RISCAL.smt2";
  public static String fileNameSMT_LIBScript=defaultFileNameSMT_LIBScript;
  
  private static final String defaultDirSMT_LIBScript=".";
  public static String dirSMT_LIBScript=defaultDirSMT_LIBScript;
  
  private static boolean defaultSingleSciptFile=true;
  public static boolean singleScriptFile=defaultSingleSciptFile;
  
  /***************************************************************************
   * If true the theorem that shall be proven is not negated.
   * This is only intended for debugging.
   **************************************************************************/
  public static boolean debug_DoNotNegateGoal=false;
  
  public static enum AuxilaryQuantfierFunctionUsage {Disabled,EnabledRespectSize,EnabledAlwaysUse};  
  private static final AuxilaryQuantfierFunctionUsage defaultQuantifierAuxFunctionMode=
  		AuxilaryQuantfierFunctionUsage.EnabledRespectSize;
  /***************************************************************************
   * Decides if in the different quantifier expansions auxiliary functions are use to represent the quantified expression.
   * Disabled
   * 	No auxiliary functions are used.
   * EnabledRespectSize
   * 	Only if the expanded string would be longer than a certain length auxiliary functions get introduced.
   * EnabledAlwaysUse
   * 	Auxiliary functions are used in every quantifier expansion.
   **************************************************************************/
  public static AuxilaryQuantfierFunctionUsage quantifierAuxFunctionMode=defaultQuantifierAuxFunctionMode;
  

  public static enum SkipMode {RetainAll,SkipUnused,RetainTheoremsAndAxioms,RetainAxioms};
  private static final SkipMode defaultSkipDeclarations=SkipMode.SkipUnused;
  
  /***************************************************************************
   * Determines if certain declarations shall be ignored.
   * RetainAll
   * 		No declarations are ignored.
   * SkipUnusedEntities
   * 		All unrelevant entities are ignored.
   * SkipTheoremsAndUnusedEntities
   * 		All unrelevant entities plus all theorems (except the theorem which shall be proven)
   * 		are ignored.
   * 
   * Motivation
   * In many cases the list of given declarations contain declarations which are not necessary to prove
   * the given theorem. In order to make the checking of theorems easier we want to skip such declarations.
   * Obviously functions/predicates/values which do not occur in the theorem or in some entity the theorem
   * relies on can be ignored (note: currently unused types will not be ignored)
   * 
   * As all theorems in the list of declarations shall hold, i.e. they shall follow from the preceding
   * declarations, we also consider the option of ignoring these theorems, as the property of such theorems
   * anyway follows from the declarations.
   **************************************************************************/
  public static SkipMode skipDeclarations=defaultSkipDeclarations;
    
  // our standard output stream
  private static PrintWriter out = Main.getOutput();
  
  /***************************************************************************
   * Timeout for the solver. (given in seconds)
   * Note that it seems like the timeout does not work with all of the supported solvers.
   **************************************************************************/
  private static final long defaultTimeout=60;
  public static long timeOut=defaultTimeout;
  
  private static final long timeOutBuffer=0;
  
  public static boolean timing=false;
  
  public static enum Solver {Boolector,Z3,Yices,CVC4};
  private static final Solver defaultUsedSolver=Solver.Yices;
  
  /***************************************************************************
   * Determines the SMT-solver which shall be applied.
   * Note that the user has to ensure that the executable of the respective solver is in the PATH.
   * Moreover the executables have to use the respective name from 
   * (boolectorCommandName,z3CommandName,yicesCommandName,cvc4CommandName)
   * This program was written for the following versions of the respective solvers:
   * Boolector: 3.1.0
   * Z3:				4.8.7
   * Yices: 		2.6.1
   * CVC4: 			1.7
   * Other version of the solvers may for example require other command line arguments and may thus
   * not work.
   **************************************************************************/
//  public static Solver usedSolver=Solver.Boolector;
  public static Solver usedSolver=defaultUsedSolver;
  
  private static final String defaultBoolectorCommandName="boolector";
  private static final String defaultZ3CommandName="z3";
  private static final String defaultYicesCommandName="yices-smt2";
  private static final String defaultCVC4CommandName="cvc4";
  
  /*
   * The names of the executables of the solvers.
   */
  public static String boolectorCommandName=defaultBoolectorCommandName;
  public static String z3CommandName=defaultZ3CommandName;
  public static String yicesCommandName=defaultYicesCommandName;
  public static String cvc4CommandName=defaultCVC4CommandName;
  
 
  private static final String defaultDir=".";
//  private static final String defaultDir="./Solvers/";
  /**
   * The directory of the solvers
   */
  public static String dirBoolector=defaultDir;
  public static String dirZ3=defaultDir;
  public static String dirYices=defaultDir;
  public static String dirCVC4=defaultDir;
  
  // if set, overrides all other path settings
  public static String execName = null;
  
  /***************************************************************************
   * Gives the commands necessary to start boolector with timeout "timeOut"
   * @param incremental If true, enable incremental mode
   * @return A list containing the words which make up the command
   **************************************************************************/
  private static List<String> getBoolectorCommand(boolean incremental)
  {
  	String cmdName = execName != null ? execName :
  	  dirBoolector+File.separator+boolectorCommandName;
  	List<String> params=new ArrayList<String>();
    params.add(cmdName);
    if(incremental)
    	params.add("--incremental");
    if(!timing)
    	params.add("--time="+Long.toString(timeOut+timeOutBuffer));
    return params;	
  }
  
  /***************************************************************************
   * Gives the commands necessary to start z3 with timeout "timeOut"
   * @param incremental If true, enable incremental mode
   * @return A list containing the words which make up the command
   **************************************************************************/
  private static List<String> getZ3Command(boolean incremental)
  {
    String cmdName = execName != null ? execName :
      dirZ3+File.separator+z3CommandName;
    List<String> params=new ArrayList<String>();
    params.add(cmdName);
    params.add("-in");
    if(!timing)
    	params.add("-T:"+Long.toString(timeOut+timeOutBuffer));
    return params;
  }
  
  /***************************************************************************
   * Gives the commands necessary to start yices with timeout "timeOut"
   * @param incremental If true, enable incremental mode
   * @return A list containing the words which make up the command
   **************************************************************************/
  private static List<String> getYicesCommand(boolean incremental)
  {
    String cmdName = execName != null ? execName :
      dirYices+File.separator+yicesCommandName;
    List<String> params=new ArrayList<String>();
    params.add(cmdName);
    if(incremental)
    	params.add("--incremental");
    if(!timing)
    	params.add("--timeout="+Long.toString(timeOut+timeOutBuffer));
    return params;
  }
  
  /***************************************************************************
   * Gives the commands necessary to start cvc4 with timeout "timeOut"
   * @param incremental If true, enable incremental mode
   * @param fmf If true, use finite model finding on bounded quantification.
   * @return A list containing the words which make up the command
   **************************************************************************/
  private static List<String> getCVC4Command(boolean incremental,boolean fmf)
  {
    String cmdName = execName != null ? execName :
      dirCVC4+File.separator+cvc4CommandName;
  	//the timeout is given in milliseconds thus *1000
    List<String> params=new ArrayList<String>();
    params.add(cmdName);
    params.add("--lang=smtlib2.6");
    if(incremental)
    	params.add("--incremental");
    if(fmf)
    	params.add("--fmf-bound");
    if(!timing)
    	params.add("--tlimit="+Long.toString((timeOut+timeOutBuffer)*1000));
    return params;
  }  
  
  /**
   * The process of the SMT-solver
   */
  private static Process smtProcess;
  
  private static Lock processLock = new ReentrantLock();
  
  private static boolean wasAborted=false;
  
  private static void resetAborted()
  {
  	wasAborted=false;
  }
  
    
  public static void terminateSolver()
  {
  	processLock.lock();
  	wasAborted=true;
  	if(smtProcess!=null)
  		smtProcess.destroy();
  	processLock.unlock();
  }
  
  public static void restoreDefaultSettings()
  {
  	usedSolver=defaultUsedSolver;
  	quantifierAuxFunctionMode=defaultQuantifierAuxFunctionMode;
  	functionDefinitionTreatmentMethod=defaultFunctionDefinitionTreatmentMethod;
  	skipDeclarations=defaultSkipDeclarations;
  	setUseSkolemisation(defaultSkolemMode);
    guardChoices=UsageModeGuardsForChoices.NoGuards;
  	setUnquantify(true);
  	checkDependencies = false;
  	enableiterativeTheoremChecking = false;
  	writeTranslatedSpecToFile=defaultWriteTranslatedSpecToFile;
  	singleScriptFile=defaultSingleSciptFile;
  	// do not reset file paths
  	// dirSMT_LIBScript=defaultDirSMT_LIBScript;
  	// fileNameSMT_LIBScript=defaultFileNameSMT_LIBScript;
  	// dirBoolector=defaultDir;
  	// dirZ3=defaultDir;
  	// dirYices=defaultDir;
  	// dirCVC4=defaultDir;	
  	// boolectorCommandName=defaultBoolectorCommandName;
  	// z3CommandName=defaultZ3CommandName;
  	// yicesCommandName=defaultYicesCommandName;
  	// cvc4CommandName=defaultCVC4CommandName;  	
  	timeOut=defaultTimeout;
  }
  
  private static void diagnosticPrints(List<Declaration> declst)
  {
	  	out.println("------------intermediate Start------------");  
	  	for(Declaration d:declst)
		  {
		  	out.print(d);
		  }
		  out.println("------------intermediate End------------");  
  }
  
  private static boolean determineUnquantificationSettings(TranslationInformation ti,boolean iterativeUsage)
  {	
  	if(unquantify)
  		return true;
  	else
  	{
  		if(usedSolver==Solver.Boolector)
  		{
  			if(ti.nonNullaryUninterpretedFunctionsNecessary)
  			{
  				out.println("WARNING: quantifiers are eliminated (uninterpreted functions are required).");
  				return true;
  			}
  			else if(iterativeUsage)
  			{
  				out.println("WARNING: quantifiers are eliminated (iterative checking is required).");
  				return true;
  			}
  			else
  			{
  				return false;
  			}  				
  		}
  		else if(usedSolver==Solver.Yices)
  		{
  			out.println("WARNING: quantifiers are eliminated (Yices does not support quantification).");
  			return true;
  		}
  		else
  		{
  			if(ti.nonNullaryUninterpretedFunctionsNecessary&&!SMT_LIB_Quantifiers_UseUFs)
  			{
  				out.println("WARNING: quantifiers are eliminated (uninterpreted functions are required).");
  				return true;
  			}
  			else
  			{
  				return false;
  			}
  		}
  	}
  }
  
  private static String addInformationToTranslation(String translation,TranslationInformation ti,boolean eliminateQuantifiers)
  {
	  translation+="(exit)";
	  if(eliminateQuantifiers)
	  {
	  	translation="(set-logic QF_UFBV)\n"+translation;
	  }
	  else if(!ti.nonNullaryUninterpretedFunctionsNecessary)
	  {
	  	translation="(set-logic BV)\n"+translation;
	  }
	  else
	  {
	  	translation="(set-logic UFBV)\n"+translation;
	  }
	  return translation;
  }
  
  private static Pair<Boolean,String> initialiseSMTProcess(TranslationInformation ti,boolean eliminateQuantifiers, 
  		boolean iterativeUsage,boolean redirectErrorStream)
  {
	  String selectedSolver;
	  ProcessBuilder procBuild=new ProcessBuilder();
	  if(redirectErrorStream)
	  {
	  	procBuild.redirectErrorStream(true);
	  }
	  switch(usedSolver)
	  {
		case Boolector:
			procBuild.command(getBoolectorCommand(iterativeUsage));
			selectedSolver="Boolector";
			break;
		case Z3:
			procBuild.command(getZ3Command(iterativeUsage));
			selectedSolver="Z3";
			break;
		case Yices:
			procBuild.command(getYicesCommand(iterativeUsage));
			selectedSolver="Yices";
			break;
		case CVC4:
			procBuild.command(getCVC4Command(iterativeUsage,!eliminateQuantifiers&&ti.nonNullaryUninterpretedFunctionsNecessary));
			selectedSolver="CVC4";
			break;
		default:
			out.println("ERROR: Invalid selection of a SMT-Solver.");
			return new Pair<Boolean, String>(false, "");
	  }
	  processLock.lock();
	  if(wasAborted)
	  {
	  	out.println("ERROR: the computation was aborted.");
	  	processLock.unlock();
	  	return new Pair<Boolean, String>(false, selectedSolver);
	  }
	  try
	  {
	  	smtProcess=procBuild.start();
	  	return new Pair<Boolean, String>(true, selectedSolver);
	  }		  
	  catch(Exception e)
	  {
	  	out.println("ERROR: the SMT solver "+selectedSolver+" could not be started.");
	  	return new Pair<Boolean, String>(false, selectedSolver);
	  }	
	  finally
	  {
	  	processLock.unlock();
	  }
  }
  
  private static void writeTranslationToFile(String translation,String theoremName)
  {
  	if(SMT.singleScriptFile)
  	{
	  	try(BufferedWriter writer = new BufferedWriter(new FileWriter(dirSMT_LIBScript+File.separator+fileNameSMT_LIBScript)))
	  	{
	  		writer.write(translation);
	  	}
	  	catch (Exception e) 
	  	{
				out.println("The translated Spec could not be written to a file.");
			}	 
  	}
  	else
  	{
  		String fileNameSuggestion=theoremName;
  		String fileName=fileNameSuggestion;
  		File f=new File(dirSMT_LIBScript, fileName);
  		int counter=1;
  		while(f.exists())
  		{
  			fileName=fileNameSuggestion+"_"+Integer.toString(counter);
  		}
  		try(BufferedWriter writer = new BufferedWriter(new FileWriter(dirSMT_LIBScript+File.separator+fileName)))
	  	{
	  		writer.write(translation);
	  	}
	  	catch (Exception e) 
	  	{
				out.println("The translated Spec could not be written to a file.");
			}	 
  	}
  }

  /***************************************************************************
   * Apply SMT solver to check the *validity* of the theorem 
   * in the context of the declarations preceding the theorem.
   * @param theorem the definition of the theorem
   * @param declarations the declarations preceding the theorem
   * @return true if the theorem is valid
   **************************************************************************/
  public static boolean check(TheoremParamDefinition theorem, 
    List<Declaration> declarations)
  {
  	resetAborted();
  	//When out was set the appropriate PrintWriter maybe was not available	
  	out=Main.getOutput();
	  String translation=null;
	  try
	  {
	  	long initialTime = System.currentTimeMillis();
	  	Pair<List<Declaration>,DeclarationPreprocessor.TranslationInformation> pre= 
	  			DeclarationPreprocessor.translate(theorem, declarations);
	  	if(diagnosticPrints)
	  		diagnosticPrints(pre.first);
	  	boolean eliminateQuantifiers=determineUnquantificationSettings(pre.second,checkDependencies);
		  translation=DeclarationTranslator.translate(pre.first,pre.second,eliminateQuantifiers);
		  long translationTerminationTime=System.currentTimeMillis();
		  if(translation==null)
		  {
		  	out.println("ERROR: An error occured in the translation to SMT-LIB");
		  	return false;		  	
		  }
		  translation=addInformationToTranslation(translation,pre.second,eliminateQuantifiers);
		  String selectedSolver;		  
		  Pair<Boolean,String> p=initialiseSMTProcess(pre.second,eliminateQuantifiers,checkDependencies,false);
		  if(!p.first)
		  	return false;
		  else
		  	selectedSolver=p.second;		  
		  try (InputStream solverIn=smtProcess.getInputStream();
		  		InputStreamReader inRead=new InputStreamReader(solverIn);
		  		BufferedReader solverReader=new BufferedReader(inRead);
		  		InputStream solverError=smtProcess.getErrorStream();
		  		InputStreamReader errorRead=new InputStreamReader(solverError);
		  		BufferedReader errorReader=new BufferedReader(errorRead);
		  		OutputStream solverOut=smtProcess.getOutputStream();
		  		PrintWriter solverWriter=new PrintWriter(solverOut,true);		  
		  		)
		  {
		  	out.println("The SMT solver " + selectedSolver + " started execution.");
		  	long solverStartTime = System.currentTimeMillis();
		  	solverWriter.println(translation);
		  	boolean finished=smtProcess.waitFor(timeOut, TimeUnit.SECONDS);
		  	long endTime = System.currentTimeMillis();
		  	if(wasAborted)
		  	{
		  		out.println("The SMT solver was terminated!");
		  		return false;
		  	}
		  	if(!finished)
		  	{
		  		smtProcess.destroy();
		  		out.println("ERROR: solver timeout.");
		  		return false;
		  	}		  		
		  	String line;
		  	List<String> output=new ArrayList<String>();
		  	List<String> errors=new ArrayList<String>();
		  	while((line=solverReader.readLine())!=null)
		  		output.add(line);
		  	while((line=errorReader.readLine())!=null)
		  		errors.add(line);
		  	if(errors.size()!=0)
		  	{
		  		if(printResult)
		  		{
		  			out.println("ERROR: the solver reported "+
		  					String.join(System.lineSeparator(), errors)+".");
		  		}
		  		return false;
		  	}
		  	String result;
	  		if(checkDependencies)
	  		{
	  			if(output.size()<2)
	  			{
	  				if(printResult)
	  				{
	  					out.println("ERROR: try without checking dependencies.");
	  				}
	  				return false;
	  			}
	  			if(!(output.get(0).equals("unsat")||output.get(0).equals("sat")))
	  			{
	  				if(printResult)
	  				{
	  					out.println("ERROR: try without checking dependencies.");
	  				}
	  				return false;
	  			}
	  			if(output.get(0).equals("unsat"))
	  			{
	  				if(printResult)
	  				{
	  				  out.println("ERROR: invalid dependencies (use option \"Full Choice Guards\").");
	  				}
	  				return false;
	  			}
	  			result=String.join("", output.subList(1, output.size()));
	  		}
	  		else
	  			result=String.join("", output);
		  	if(result.equals("unsat")||result.equals("sat"))
		  	{
		  		if(printResult)
		  		{
		  			long translationTime=translationTerminationTime - initialTime;
	  				long solverTime=endTime - solverStartTime;
	  				long totalTime=endTime - initialTime;
	  				if(result.equals("unsat"))
	  				{
	  					out.println("Theorem is valid ("+Long.toString(totalTime)+
	  							" ms, translation: "+Long.toString(translationTime)+" ms, decision: "+Long.toString(solverTime)
	  							+" ms).");
	  					return true;
	  				}
	  				else
	  				{
	  					out.println("ERROR: theorem is invalid ("+Long.toString(totalTime)+
	  							" ms, translation: "+Long.toString(translationTime)+" ms, decision: "+Long.toString(solverTime)
	  							+" ms).");
	  					return false;
	  				}
		  		}
		  		if(result.equals("unsat"))
		  			return true;
		  		else 
		  			return false;
		  	}
		  	else if(result.toLowerCase().contains("unknown"))
		  	{
		  		if(printResult)
		  		{
		  			out.println("ERROR: The solver reports unknown.");
		  		}
		  		return false;
		  	}
		  	else if(result.toLowerCase().contains("timeout"))
		  	{		  		
		  		if(printResult)
		  		{
		  			out.println("ERROR: solver timeout.");
		  		}
		  		return false;
		  	}
		  	else
		  	{
		  		if(printResult)
		  		{
		  			out.println("ERROR: the solver reported "+result+".");
		  		}
		  		return false;
		  	}		  			  			  		
		  } 
		  catch (InterruptedException e) 
		  {
		  	if(wasAborted)
		  	{
		  		out.println("The SMT-Solver was terminated!");
		  		return false;
		  	}
				e.printStackTrace();
				out.println("Error: InterruptedException");
				return false;
			}
		  finally 
		  {
				smtProcess.destroy();
			  if(writeTranslatedSpecToFile)
			  {
			  	writeTranslationToFile(translation,theorem.ident.string);
			  }
			}		  
	  }
	  catch(NotImplemented e)
	  {
	  	out.println("ERROR: Not implemented RISCAL component detected");
	  	out.println(e.toString());
	  	if(e.tree!=null)
	  	{
	  		out.println("Error detected in:");
	  		out.println(e.tree.toString());  	
	  	}	
	  	help();
	  	return false;
	  }
	  catch(IlligalUsage e)
	  {
	  	out.println("ERROR: Internal error -  Some invariant was violated.");
	  	out.println(e.toString());
	  	if(e.tree!=null)
	  	{
	  		out.println("Error detected in:");
	  		out.println(e.tree.toString());  	
	  	}
	  	help();
	  	return false;
	  }
	  catch(TooLargeTypeSize e)
	  {
	  	out.println("ERROR: An used type does not obey the size limit of Types.");
	  	String s=e.type==null?"":e.type.toString();
	  	out.println("A type must not have a size larger than "+Long.toString(maximalTypeSize)+
	  			" but the type "+s+" has length "+Long.toString(e.usedSize));
	  	return false;
	  }
	  catch(OutOfMemoryError e)
	  {
	  	out.println("ERROR: Out of Memory "
	  			+ "The given specification could not be translated."
	  			+ " To performe the translation more memory would be required.");
	  	return false;
	  }
	  catch(IOException e)
	  {
	  	out.println("ERROR: a problem with the communication with the solver occured.");
	  	return false;
	  }
	  finally 
	  {
	  	out.flush();
		}
  }
   
  
  static class ResultPrinter extends Thread
  {
  	private BufferedReader reader;
  	private List<TheoremParamDefinition> thms;
  	public volatile List<Boolean> results;
  	private List<Consumer<Boolean>> handlers;
  	public ResultPrinter(BufferedReader r,List<TheoremParamDefinition> t,
  	  List<Consumer<Boolean>> h)
  	{
  		reader=r;  		
  		thms=t;
  		handlers=h;
  		results=new ArrayList<Boolean>();
  	}
  	
    public void run(){
      try 
      {
      	String line;
      	ListIterator<TheoremParamDefinition> it=thms.listIterator();
      	int i = 0;
      	while((line=reader.readLine())!=null)
      	{
          Consumer<Boolean> handler = handlers == null ? null : handlers.get(i); i++;
      		if(!it.hasNext())
      			return;
      		String theoremName=it.next().ident.string;
      		if(checkDependencies)
      		{
      			if(line.equals("sat"))
      			{
      				line=reader.readLine();
      				if(line!=null)
      				{
      					if(line.equals("sat"))
      					{
      						out.println("Theorem "+theoremName+" is invalid.");
      						results.add(false);
      						if (handler != null) handler.accept(false);
      					}
      					else if(line.equals("unsat"))
      					{
      						out.println("Theorem "+theoremName+" is valid.");
      						results.add(true);
                  if (handler != null) handler.accept(true);
      					}
      					else if(line.equals("unknown"))
      					{
      						out.println("The selected solver could not conclude a result for "+theoremName);
      						results.add(false);
                  if (handler != null) handler.accept(false);
      					}
      					else
      					{
      						out.println("Error: "+line);
      						out.println("The computation is aborted!");
      						results.add(false);
      						terminateSolver();
      						return;
      					}
      				}
      			}
      			else if(line.equals("unsat"))
      			{
      				out.println("Dependencies of "+theoremName+" are violated.");
      				results.add(false);
      			}
      			else if(line.equals("unknown"))
      			{
      				line=reader.readLine();
      			}
      			else
      			{
      				out.println("Error: "+line);
  						out.println("The computation is aborted!");
  						results.add(false);
  						terminateSolver();
  						return;
      			}
      		}
      		else
      		{
      			if(line.equals("sat"))
  					{
  						out.println("Theorem "+theoremName+" is invalid.");
  						results.add(false);
              if (handler != null) handler.accept(false);
  					}
  					else if(line.equals("unsat"))
  					{
  						out.println("Theorem "+theoremName+" is valid.");
  						results.add(true);
              if (handler != null) handler.accept(true);
  					}
  					else if(line.equals("unknown"))
  					{
  						out.println("The selected solver could not conclude a result for "+theoremName);
  						results.add(false);
              if (handler != null) handler.accept(false);
  					}
  					else
  					{
  						out.println("Error: "+line);
  						out.println("The computation is aborted!");
  						results.add(false);
  						terminateSolver();
  						return;
  					}
      		}
      	}
			} catch (IOException e) 
      {	
			}
    }
  }
  
  /**
   * Checks multiple theorems
   * @param theorems The theorems that shall be checked
   * @param declarations The declaration the theorems may depend on
   * @param handlers to be notified about the results of the theorems (may be null or contain null elements)
   * @return An element of this list is true iff the theorem in theorems at the corresponding position is valid
   */
  public static List<Boolean> checkMultipleTheorems(List<TheoremParamDefinition> theorems, 
      List<Declaration> declarations, List<Consumer<Boolean>> handlers)
  {
  	if(theorems.size()==0)
  		return new ArrayList<Boolean>();
  	else if(theorems.size()==1)
  	{
  		boolean r=check(theorems.get(0),declarations);
  		List<Boolean> results=new ArrayList<Boolean>(1);
  		results.add(r);
  		Consumer<Boolean> handler = handlers == null ? null : handlers.get(0);
  		if (handler != null) handler.accept(r);
  		return results;
  	} 	
  	resetAborted();
  	//When out was set the appropriate PrintWriter maybe was not available	
  	out=Main.getOutput();
	  String translation=null;	  
	  try
	  {
	  	long initialTime = System.currentTimeMillis();
	  	Triple<List<Declaration>, List<List<Declaration>>, TranslationInformation> pre=
	  			DeclarationPreprocessor.translateMultiple(theorems, declarations);
	  	boolean eliminateQuantifiers=determineUnquantificationSettings(pre.third,true);
	  	translation=DeclarationTranslator.translateMultiple(pre.first, pre.second, pre.third, eliminateQuantifiers);
		  long translationTerminationTime=System.currentTimeMillis();
		  if(translation==null)
		  {
		  	out.println("ERROR: An error occured in the translation to SMT-LIB");
		  	List<Boolean> lst=Collections.nCopies(theorems.size(), false);
		  	return lst;		  	
		  }
		  translation=addInformationToTranslation(translation,pre.third,eliminateQuantifiers);
		  String selectedSolver;		  
		  Pair<Boolean,String> p=initialiseSMTProcess(pre.third,eliminateQuantifiers,true,true);
		  if(!p.first)
		  {
		  	List<Boolean> lst=Collections.nCopies(theorems.size(), false);
		  	return lst;		
		  }
		  else
		  	selectedSolver=p.second;
		  try (InputStream solverIn=smtProcess.getInputStream();
		  		InputStreamReader inRead=new InputStreamReader(solverIn);
		  		BufferedReader solverReader=new BufferedReader(inRead);
		  		OutputStream solverOut=smtProcess.getOutputStream();
		  		PrintWriter solverWriter=new PrintWriter(solverOut,true);		  
		  		)
		  {
		  	out.println("The SMT solver " + selectedSolver + " started execution.");
		  	long solverStartTime = System.currentTimeMillis();
		  	ResultPrinter rP=new ResultPrinter(solverReader, theorems, handlers);
		  	rP.start();
		  	solverWriter.println(translation);
		  	boolean finished=smtProcess.waitFor(timeOut, TimeUnit.SECONDS);
		  	long endTime = System.currentTimeMillis();
		  	rP.join();
		  	if(wasAborted)
		  	{
		  		out.println("The SMT solver was terminated!");		
		  	}
		  	if(!finished)
		  	{
		  		smtProcess.destroy();
		  		out.println("ERROR: solver timeout.");
		  	}
		  	if(printResult)
		  	{
		  		long translationTime=translationTerminationTime - initialTime;
  				long solverTime=endTime - solverStartTime;
  				long totalTime=endTime - initialTime;
  				out.println("Total time: "+Long.toString(totalTime)+" ms, translation: "
  						+Long.toString(translationTime)+" ms, decision: "+Long.toString(solverTime)+" ms.");
		  	}
		  	List<Boolean> results=rP.results;
		  	if(results.size()<theorems.size())
		  	{
		  		results.addAll(Collections.nCopies(theorems.size()-results.size(), false));
		  	}
		  	return results; 		  	
		  } 
		  catch (InterruptedException e) 
		  {
		  	if(wasAborted)
		  	{
		  		out.println("The SMT-Solver was terminated!");
		  		return Collections.nCopies(theorems.size(), false);
		  	}
				e.printStackTrace();
				out.println("Error: InterruptedException");
				return Collections.nCopies(theorems.size(), false);
			}
		  finally 
		  {
				smtProcess.destroy();
			  if(writeTranslatedSpecToFile)
			  {
			  	writeTranslationToFile(translation,theorems.get(0).ident.string);
			  }
		  }
	  }
	  catch(NotImplemented e)
	  {
	  	out.println("ERROR: Not implemented RISCAL component detected");
	  	out.println(e.toString());
	  	if(e.tree!=null)
	  	{
	  		out.println("Error detected in:");
	  		out.println(e.tree.toString());  	
	  	}	
	  	help();
	  	return Collections.nCopies(theorems.size(), false);
	  }
	  catch(IlligalUsage e)
	  {
	  	out.println("ERROR: Internal error -  Some invariant was violated.");
	  	out.println(e.toString());
	  	if(e.tree!=null)
	  	{
	  		out.println("Error detected in:");
	  		out.println(e.tree.toString());  	
	  	}
	  	help();
	  	return Collections.nCopies(theorems.size(), false);
	  }
	  catch(TooLargeTypeSize e)
	  {
	  	out.println("ERROR: An used type does not obey the size limit of Types.");
	  	String s=e.type==null?"":e.type.toString();
	  	out.println("A type must not have a size larger than "+Long.toString(maximalTypeSize)+
	  			" but the type "+s+" has length "+Long.toString(e.usedSize));
	  	return Collections.nCopies(theorems.size(), false);
	  }
	  catch(OutOfMemoryError e)
	  {
	  	out.println("ERROR: Out of Memory "
	  			+ "The given specification could not be translated."
	  			+ " To performe the translation more memory would be required.");
	  	return Collections.nCopies(theorems.size(), false);
	  }
	  catch(IOException e)
	  {
	  	out.println("ERROR: a problem with the communication with the solver occured.");
	  	return Collections.nCopies(theorems.size(), false);
	  }
	  finally 
	  {
	  	out.flush();
		}	  
  }
  
  
  /***************************************************************************
   * Writes some notes on the usage of the SMT-LIB translator to the PrintWriter out.
   **************************************************************************/
  private static void help()
  {
  	String limitations="- Recursive types are not supported"+System.lineSeparator()+
  			"- Supported Type Sizes:"+System.lineSeparator()+
  			"    + Theoretical upper limit: 2^31-1"+System.lineSeparator()+
  			"    + In practice the limit can be lower than 2^31-1, depending on the available memory"+System.lineSeparator()+
  			"- Theorems must not be used like predicates / values"+System.lineSeparator()+
  			"- If a Theorem shall be analysed the user is resposible to ensure that the necessary proof obligations "
  			+ "hold. If this is not the case, results may be incorrect. "
  			+ "Thus the preconditions of called functions must hold"
  			+ "indexes of arrays must not be out of bounds,...";
  	out.println("------------------------------------------------");
  	out.println("Limitations of the Translation:");
  	out.println(limitations);
  	out.println("------------------------------------------------");
  }
  
  
  
  
  /***************************************************************************
   * Throws a TooLargeTypeSize exception.
   * @param msg The message of the exception.
   * @param size The size of the type that causes the exception.
   * @param t The type that causes the exception.
   **************************************************************************/
  public static void tooLargeType(String msg,long size,Type t)
  {
  	throw new TooLargeTypeSize(msg, t, size);
  }
  
  /***************************************************************************
   * The exception raised if a type exceeds the limit on type sizes.
   **************************************************************************/
  public static final class TooLargeTypeSize extends RuntimeException
  {
		private static final long serialVersionUID = 8168386009251108979L;
		
		private long usedSize;
		private Type type;
		
		/***************************************************************************
		 * Create a too large type exception with the denoted message originating from 
		 * the denoted type.
		 * @param msg the message
		 * @param t The type that caused the exception.
		 * @param usedSize The size of the type.
		 **************************************************************************/
		public TooLargeTypeSize(String msg,Type t,long usedSize)
		{
			super(msg);
			this.usedSize=usedSize;
			this.type=t;
		} 	
  }
  
  
  
  /***************************************************************************
   * Throw a not implemented error.
   * @param tree the phrase causing the error
   * @param msg the message reported by the error
   **************************************************************************/
  public static void notImplemented(AST tree, String msg)
  {
    throw new NotImplemented(tree, msg);
  }
  
  /***************************************************************************
   * The exception is raised if a not yet treated 
   * component of the RISCAL language is discovered.
   **************************************************************************/
  public static final class NotImplemented extends RuntimeException
  {
  	private static final long serialVersionUID = -7542749988510143600L;
    public final AST tree;
    
    /*************************************************************************
     * Create a NotImplemented Exception.
     * @param tree the abstract syntax tree of the phrase
     * @param msg the message
     ************************************************************************/
    public NotImplemented(AST tree, String msg) 
    { 
      super(msg); 
      this.tree = tree;
    }
  }
  
  
  /***************************************************************************
   * Throw a illegal usage error.
   * @param tree the phrase causing the error
   * @param msg the message reported by the error
   **************************************************************************/
  public static void illegalUsage(AST tree, String msg)
  {
    throw new IlligalUsage(tree, msg);
  }
  
  /***************************************************************************
   * The exception is raised if a not some invariant is violated.
   * Thus this exception serves mainly as a debug exception and should
   * not be raised otherwise.
   **************************************************************************/
  public static final class IlligalUsage extends RuntimeException
  {
		private static final long serialVersionUID = -3812126517811045796L;
		public final AST tree;
		
		/*************************************************************************
     * Create a IlligalUsage Exception.
     * @param tree the abstract syntax tree of the phrase
     * @param msg the message
     ************************************************************************/
		public IlligalUsage(AST tree, String msg)
		{
			super(msg);
			this.tree=tree;
		}
  	
  } 
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------