// ---------------------------------------------------------------------------
// SetTranslator.java
// A class for the translation of various set expressions to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: SetTranslator.java,v 1.6 2020/06/23 05:44:15 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.StringJoiner;
import java.util.function.Function;
import java.util.stream.Collectors;

import riscal.semantics.Types.Value;
import riscal.semantics.Types.Value.Set;
import riscal.semantics.Values;
import riscal.smt.ExpressionTranslator.TypeEnumerationInformation;
import riscal.smt.ExpressionTranslator.ValueContainer;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.BigIntersectExp;
import riscal.syntax.AST.Expression.BigUnionExp;
import riscal.syntax.AST.Expression.CartesianExp;
import riscal.syntax.AST.Expression.EmptySetExp;
import riscal.syntax.AST.Expression.EnumeratedSetExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.InSetExp;
import riscal.syntax.AST.Expression.IntersectExp;
import riscal.syntax.AST.Expression.IntervalExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.Powerset1Exp;
import riscal.syntax.AST.Expression.Powerset2Exp;
import riscal.syntax.AST.Expression.PowersetExp;
import riscal.syntax.AST.Expression.SetBuilderExp;
import riscal.syntax.AST.Expression.SetSizeExp;
import riscal.syntax.AST.Expression.SubsetExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.Expression.UnionExp;
import riscal.syntax.AST.Expression.WithoutExp;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.MapType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.AST.Type.UnitType;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.util.Seq;
import riscal.util.Seq.Next;

public class SetTranslator extends ASTVisitor.Base<String>
{
	/*
	 * Suggestions for the names of functions representing various kinds
	 * of set related expressions.
	 */
	private static final String defaultTypeEnumerationName="setEnum";
	private static final String defaultCardinalityName="card";
	private static final String defaultCardinalityFilterName="cardFilter";
	private static final String defaultPowerSetFunctionName="powerSet";
	private static final String defaultBigUnionName="bigUnion";
	private static final String defaultBigIntersectionName="bigIntersection";
	
	/**
	 * The underlying ExpressionTranslator
	 */
	private ExpressionTranslator expTrans;
	/**
	 * The SetType that is used for the current translation.
	 */
	private SetType usedType;
	/**
	 * The bit vector length that is used for the current translation.
	 */
	private int setVectorLength;
	/**
	 * Used to process values of the current usedType.
	 */
	private ValueContainer valC;
	
	/**
	 * Contains the names of the already generated functions used to
	 * represent cardinality expressions.
	 * The map assigns the name of the cardinality function to the
	 * number of elements of the underlying type of a set type.
	 */
	private HashMap<Integer,String> cardinalityFunctionNames=new HashMap<Integer, String>();

	/**
	 * Entries shall only contain SetTypes as keys.
	 * Assigns to a type T the name of a function that takes the bit vector
	 * representations of 0<=x<=|T| and returns a bit vector representing the
	 * set that contains all sets (of the respective type) of cardinality x.
	 * 
	 * Improvement:
	 * The keys of the map could be integers representing the the cardinality
	 * of the respective set universe.
	 */
	private HashMap<Type,String> filtersForElementsOfCardinalityN=new HashMap<Type, String>();
	/**
	 * The map assigns the name of the function used to represent power set
	 * expressions to the number of elements of the underlying type.
	 */
	private HashMap<Integer,String> powerSetFunctions=new HashMap<Integer, String>();
	/**
	 * The map assigns the name of the function used to represent bigunion
	 * expressions to the number of elements of the underlying type.
	 */
	private HashMap<Integer, String> bigUnion=new HashMap<Integer, String>();
	/**
	 * The map assigns the name of the function used to represent bigintersection
	 * expressions to the number of elements of the underlying type.
	 */
	private HashMap<Integer, String> bigIntersection=new HashMap<Integer, String>();
	
	/***************************************************************************
	 * Create a SetTranslator with the given underlying ExpressionTranslator.
	 * @param et The underlying ExpressionTranslator.
	 **************************************************************************/
	public SetTranslator(ExpressionTranslator et)
	{
		this.expTrans=et;
	}
	
	/***************************************************************************
	 * Compute the size of the universe of the type of the given expression.
	 * @param exp A set expression.
	 * @return The size of the underlying type of the type of the given expression.
	 **************************************************************************/
	private static int getLength(Expression exp)
	{
		SetType t=(SetType)Utilities.getBaseType(exp.getType());
		return ExpressionTranslator.getTypeSize(t.type);		
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * @param exp The expression that shall be translated.
	 * Has to be a set related expression.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp)
	{
		SetType oldUsedType=usedType;
		usedType=(SetType)getMaximalSet(exp);
		int len=setVectorLength;
		setVectorLength=ExpressionTranslator.getTypeSize(usedType.type);
		ValueContainer aux=valC;
		valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, usedType);
		String str=exp.accept(this);
		if(exp.getType() instanceof SetType)
		{
			str=decreaseSet(str, exp.getType(), usedType);
		}
		valC=aux;
		usedType=oldUsedType;
		setVectorLength=len;
		return str;		
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * The expression is treated like it had the type type.
	 * @param exp The expression that shall be translated.
	 * @param type The possible values of exp shall be representable
	 * by this type.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp,Type type)
	{
		SetType oldUsedType=usedType;
		usedType=(SetType)getMaximalSet(exp);
		usedType=(SetType) determineCommonSuperType(usedType, type);
		int len=setVectorLength;
		setVectorLength=ExpressionTranslator.getTypeSize(usedType.type);
		ValueContainer aux=valC;
		valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, usedType);
		String str=exp.accept(this);
		if(exp.getType() instanceof SetType)
		{
			str=decreaseSet(str, type, usedType);
		}
		valC=aux;
		usedType=oldUsedType;
		setVectorLength=len;
		return str;		
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to oldType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to newType.
	 * Condition:
	 * oldType must be contained in newType
	 * @param s The SMT-LIB expression that shall be transformed.
	 * @param oldType The type that corresponds to the current sort.
	 * @param newType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * newType
	 **************************************************************************/
	String adaptTranslation(String s,SetType oldType,SetType newType)
	{
		return enlargeSet(s, oldType, newType);
	}
	
	/***************************************************************************
	 * Compute a SetType covering all SetType occurring in the given expression.
	 * @param exp The expression of interest
	 * @return A Type fulfilling the above description.
	 **************************************************************************/
	private Type getMaximalSet(Expression exp)
	{
		//Predicates
		if(exp instanceof EqualsExp)
		{
			EqualsExp e=(EqualsExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		else if(exp instanceof NotEqualsExp)
		{
			NotEqualsExp e=(NotEqualsExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		else if(exp instanceof SubsetExp)
		{
			SubsetExp e=(SubsetExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		else if(exp instanceof InSetExp)
		{
			InSetExp e=(InSetExp)exp;
			SetType sT=new SetType(e.exp1.getType());
			sT.setCanonical();
			Type t1=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, sT);
		}
		
		//common set operations
		else if(exp instanceof SetSizeExp)
		{
			SetSizeExp e=(SetSizeExp)exp;
			return determineCommonSuperType(e.exp.getType(), getMaximalSet(e.exp));
		}
		else if(exp instanceof UnionExp)
		{
			UnionExp e=(UnionExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		else if(exp instanceof IntersectExp)
		{
			IntersectExp e=(IntersectExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		else if(exp instanceof WithoutExp)
		{
			WithoutExp e=(WithoutExp)exp;
			Type t1=determineCommonSuperType(e.exp1.getType(), getMaximalSet(e.exp1));
			Type t2=determineCommonSuperType(e.exp2.getType(), getMaximalSet(e.exp2));
			return determineCommonSuperType(t1, t2);
		}
		
		else if(exp instanceof IfThenElseExp)
		{
			IfThenElseExp e=(IfThenElseExp)exp;
			Type t1=getMaximalSet(e.exp2);
			Type t2=getMaximalSet(e.exp3);
			Type t3=determineCommonSuperType(t1, t2);
			return t3;			
		}
		else if(exp instanceof LetParExp)
		{
			LetParExp e=(LetParExp)exp;
			return determineCommonSuperType(e.exp.getType(), getMaximalSet(e.exp));
		}
		
		SetType sT=(SetType)exp.getType();
		return sT;
	}
	
	/***************************************************************************
	 * Determines a common super type for the given types.
	 * Condition: 
	 * 		- The types have to be canonical
	 * 		- There has to be a common super type for type1 and type2
	 * 			i.e. there is a type that contains all the values of
	 * 			type1 and type2.
	 * @param type1 A type of interest. Shall be a canonical type.
	 * @param type2 A type of interest. Shall be a canonical type.
	 * @return A common super type of type1 and type2.
	 **************************************************************************/
	private static Type determineCommonSuperType(Type type1,Type type2)
	{
		return ExpressionTranslator.determineCommonSuperType(type1, type2);
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to smallType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to largeType.
	 * Condition:
	 * smallType must be contained in largeType
	 * @param set The SMT-LIB expression that shall be transformed.
	 * @param smallType The type that corresponds to the current sort.
	 * @param largeType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * largeType
	 **************************************************************************/
	private String enlargeSet(String set,Type smallType,SetType largeType)
	{
		if(smallType.equals(largeType))
			return set;
		smallType=Utilities.getBaseType(smallType);
		Type smallElementType=((SetType)smallType).type;
		Type largeElementType=largeType.type;
		if(smallElementType instanceof UnitType||smallElementType instanceof BoolType)
			return set;
		if(expTrans.giveVectorLength(smallType)==expTrans.giveVectorLength(largeType))
		{
			return set;
		}
		if(smallElementType instanceof IntType)
		{
			IntType largeElemType=(IntType)largeElementType;
			IntType smallElemType=(IntType)smallElementType;
			long lowerOffset=smallElemType.ivalue1-largeElemType.ivalue1;
			long upperOffSet=largeElemType.ivalue2-smallElemType.ivalue2;
			if(lowerOffset==0&&upperOffSet==0)
			{
				return set;
			}
			String res="((_ zero_extend "+Long.toString(lowerOffset+upperOffSet)+")"+set+")";
			long len=ExpressionTranslator.getTypeSize(smallElemType);
			String aux=IntTranslator.generateSMT_LIB_Constant(lowerOffset, lowerOffset+upperOffSet+len);
			res="(bvshl "+res+" "+aux+")";
			return res;
		}
		int[] gaps=determineGapsInSetBitVector(largeElementType, smallElementType);		
		int firstgap=0;
		for(int k=0;k<gaps.length;k++)
		{
			if(gaps[k]!=0)
			{
				firstgap=k;
				break;
			}					
			if(k==gaps.length-1)
				return set;
		}
		String res;
		if(firstgap==0)
		{
			res="#b"+"0".repeat(gaps[0]);
		}
		else
		{
			res="((_ extract "+Long.toString(firstgap-1)+" "+Long.toString(0)+")"+set+")";
			res="((_ zero_extend "+Long.toString(gaps[firstgap])+")"+res+")";
		}		
		int lastgap=firstgap;
		for(int k=firstgap+1;k<gaps.length;k++)
		{
			if(gaps[k]!=0)
			{
				String aux="((_ extract "+Long.toString(k-1)+" "+Long.toString(lastgap)+")"+set+")";
				aux="((_ zero_extend "+Long.toString(gaps[k])+")"+aux+")";
				res="(concat "+aux+" "+res+")";
				lastgap=k;
			}
		}
		if(lastgap!=gaps.length-1)
		{
			String aux="((_ extract "+Long.toString(gaps.length-2)+" "+Long.toString(lastgap)+")"+set+")";
			res="(concat "+aux+" "+res+")";
		}
		return res;
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to largeType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to smallType.
	 * Condition:
	 * smallType must be contained in largeType
	 * @param set The SMT-LIB expression that shall be transformed.
	 * @param smallType The type that corresponds to the current sort.
	 * @param largeType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * smallType
	 **************************************************************************/
	private String decreaseSet(String set,Type smallType,SetType largeType)
	{
		if(smallType.equals(largeType))
			return set;
		smallType=Utilities.getBaseType(smallType);
		Type smallElementType=((SetType)smallType).type;
		Type largeElementType=largeType.type;
		if(smallElementType instanceof UnitType||smallElementType instanceof BoolType)
			return set;
		if(expTrans.giveVectorLength(smallType)==expTrans.giveVectorLength(largeType))
		{
			return set;
		}
		if(smallElementType instanceof IntType)
		{
			IntType largeElemType=(IntType)largeElementType;
			IntType smallElemType=(IntType)smallElementType;
			long lowerOffset=smallElemType.ivalue1-largeElemType.ivalue1;
			long upperOffSet=largeElemType.ivalue2-smallElemType.ivalue2;
			if(lowerOffset==0&&upperOffSet==0)
				return set;
			long len=ExpressionTranslator.getTypeSize(largeElemType);
			String res="((_ extract "+Long.toString(len-1-upperOffSet)+" "+Long.toString(lowerOffset)+")"+set+")";
			return res;			
		}
		int[] gaps=determineGapsInSetBitVector(largeElementType, smallElementType);	
		
		int startAt=0;
		if(gaps[0]!=0)
		{
			startAt=gaps[0];
		}
		int firstGap=0;
		for(int k=1;k<gaps.length;k++)
		{
			if(gaps[k]!=0)
			{
				firstGap=k;
				break;
			}					
			if(k==gaps.length-1)
			{
				if(startAt==0)
					return set;
				else
				{
					return "((_ extract "+Long.toString(startAt+k-1)+" "+Long.toString(startAt)+")"+set+")";		
				}
			}
		}
		int finishAt=firstGap+startAt;
		String res="((_ extract "+Long.toString(finishAt-1)+" "+Long.toString(startAt)+")"+set+")";		
		startAt=finishAt+gaps[firstGap];
		int len=0;
		for(int i=firstGap+1;i<gaps.length;i++)
		{
			if(gaps[i]==0)
			{
				len++;			
			}
			else
			{
				res="(concat "+"((_ extract "+Long.toString(startAt+len)+" "+Long.toString(startAt)+")"+set+")"+res+")";
				startAt+=len+gaps[i]+1;
				len=0;
			}
		}
		if(len!=0)
		{
			res="(concat "+"((_ extract "+Long.toString(startAt+len-1)+" "+Long.toString(startAt)+")"+set+")"+res+")";
		}
		return res;
	}
	
	/***************************************************************************
	 * Compute an array of length |containedType|+1
	 * In the following we say that a value is smaller than another one
	 * if it appears before the other value in Values.getValueSeq(type)
	 * Similarly we say that a value is larger than another one if it appears
	 * after the other value Values.getValueSeq(type).
	 * Moreover we denote the ith element of a type as the ith element of
	 * Values.getValueSeq(type) (we start counting with 1)
	 * The first(=[0]) element of the array contains the number of elements of superType
	 * that are smaller than the first element of superType.
	 * The last element of the array contains the number of elements of superType
	 * that are larger than the last element of containedType.
	 * For 1<index<result.length-1 result[i] contains the number of elements 
	 * of superType that are larger than the ith element of 
	 * containedType and smaller than the (i+1)th element of containedType.
	 * 
	 * Idea:
	 * Assume we want to represent {2,3} u {1,4}
	 * We can represent {2,3} by 11 and {1,4} by 1001
	 * Set containedType=SetType[IntType[2,3]] and superType=SetType[IntType[1,4]]
	 * Applying determineGapsInSetBitVector to these types gives
	 * {1,0,0,1} (=gaps)
	 * By using gaps we know that we have to append and prepend a zero to 11
	 * such that we get the representation of {2,3} with respect to superType.
	 * => bvor(0110,1001)
	 * 
	 * Conditions:
	 * 		containedType must be contained by superType
	 * 		superType shall not be a SubType
	 * 		The types must be canonical
	 * 		The containedType must have less elements than 2^31-1
	 * @param superType The contained type of interest.
	 * @param containedType The super type of interest.
	 * @return An array of length |containedType|+1
	 **************************************************************************/
	public static int[] determineGapsInSetBitVector(Type superType,Type containedType)
	{
		if(superType.equals(containedType))
		{
			int typeSize=ExpressionTranslator.getTypeSizeConsiderBuffer(containedType, 1);
			return new int[typeSize+1];
		}
		Type containedBase=Utilities.getBaseType(containedType);
		if(containedBase instanceof UnitType)
		{
			int[] res= {0,0};
			return res;
		}
		if(containedBase instanceof BoolType)
		{
			if(containedType instanceof SubType)
			{
				return determineGapsInSetBitVector_SubTypes((SubType) containedType);
			}
			int[] res= {0,0,0};
			return res;
		}
		if(containedBase instanceof IntType)
		{
			IntType iT1=(IntType)superType;
			IntType iT2=(IntType)containedBase;
			int val1=(int)(iT2.ivalue1-iT1.ivalue1);
			int val2=(int)(iT1.ivalue2-iT2.ivalue2);
			if(containedType instanceof SubType)
			{
				int[] res=determineGapsInSetBitVector_SubTypes((SubType)containedType);
				res[0]+=val1;
				res[res.length-1]+=val2;
				return res;
			}
			else
			{
				int typeSize=ExpressionTranslator.getTypeSizeConsiderBuffer(containedType, 1);
				int[] res=new int[typeSize+1];
				res[0]=val1;
				res[typeSize]=val2;
				return res;
			}			
		}
		int[] gapsInBase;
		if(containedBase instanceof TupleType)
		{
			gapsInBase=determineGapsInSetBitVector_Tuples((TupleType)superType,(TupleType)containedBase);
		}
		else if(containedBase instanceof MapType)
		{
			gapsInBase=determineGapsInSetBitVector_Maps((MapType)superType, (MapType) containedBase);
		}
		else if(containedBase instanceof SetType)
		{
			gapsInBase=determineGapsInSetBitVector_Sets((SetType)superType, (SetType)containedBase);
		}
		else
		{
			assert(false);
			return null;
		}
		if(containedType instanceof SubType)
		{
			int[] gapsInSub=determineGapsInSetBitVector_SubTypes((SubType)containedType);
			for(int i=0,baseCounter=0;i<gapsInSub.length;i++,baseCounter++)
			{
				if(gapsInSub[i]==0)
				{
					gapsInSub[i]=gapsInBase[baseCounter];
				}
				else
				{
					int limit=baseCounter+gapsInSub[i];
					for(;baseCounter<limit;baseCounter++)
					{
						gapsInSub[i]+=gapsInBase[baseCounter];
					}
					//we take this out of the loop in order to avoid an increment of basecounter in the last iteration.
					gapsInSub[i]+=gapsInBase[baseCounter];
				}
			}
			return gapsInSub;
		}
		else
		{
			return gapsInBase;
		}
	}
	
	/***************************************************************************
	 * For the property of the result see the description of determineGapsInSetBitVector.
	 * @param superType The contained type of interest.
	 * @param containedType The super type of interest.
	 * @return An array of length |containedType|+1
	 **************************************************************************/
	private static int[] determineGapsInSetBitVector_Tuples(TupleType superType,TupleType containedType)
	{
		int typeSize=ExpressionTranslator.getTypeSizeConsiderBuffer(containedType, 1);
		int[] res=new int[typeSize+1];
		int numComps=superType.types.length;
		List<int[]> gaps=new ArrayList<int[]>(numComps);
		int[] componentTypeSizes=new int[numComps];
		int[] accumulatedSizes=new int[numComps];
		for(int i=0;i<numComps;i++)
		{
			int[] b=determineGapsInSetBitVector(superType.types[i], containedType.types[i]);
			componentTypeSizes[i]=b.length-1;
			if(i==0)
			{
				accumulatedSizes[i]=1;
			}
			else
			{
				accumulatedSizes[i]=accumulatedSizes[i-1]*ExpressionTranslator.getTypeSize(superType.types[i-1]);
			}
			gaps.add(b);
		}
		int lowerOffset=0;
		for(int j=0;j<numComps;j++)
		{
			lowerOffset+=accumulatedSizes[j]*gaps.get(j)[0];
		}
		res[0]=lowerOffset;
		int[] subIndices=new int[numComps];
		for(int i=0;i<typeSize;i++)
		{
			int gapLength=0;
			if(i==typeSize-1)
			{
				for(int k=0;k<numComps;k++)
				{
					if(k<numComps-1)
					{
						gapLength+=gaps.get(k)[subIndices[k]+1]*accumulatedSizes[k];
						gapLength+=gaps.get(k+1)[subIndices[k+1]+1]*accumulatedSizes[k+1];
					}
				}
			}
			else
			{
				for(int k=0;k<numComps;k++)
				{
					if(subIndices[k]==componentTypeSizes[k]-1)
					{
						if(k<numComps-1)
						{
							gapLength+=gaps.get(k)[subIndices[k]+1]*accumulatedSizes[k];
							gapLength+=gaps.get(k)[0]*accumulatedSizes[k];
							gapLength+=gaps.get(k+1)[subIndices[k+1]+1]*accumulatedSizes[k+1];
						}
						subIndices[k]=0;
					}
					else
					{
						if(k==0)
						{
							gapLength+=gaps.get(0)[subIndices[0]+1];
						}
						subIndices[k]=subIndices[k]+1;
						break;
					}
				}
			}
			res[i+1]=gapLength;
		}
		return res;
	}
	
	/***************************************************************************
	 * For the property of the result see the description of determineGapsInSetBitVector.
	 * @param superType The contained type of interest.
	 * @param containedType The super type of interest.
	 * @return An array of length |containedType|+1
	 **************************************************************************/
	private static int[] determineGapsInSetBitVector_Maps(MapType superType,MapType containedType)
	{
		int typeSize=ExpressionTranslator.getTypeSize(containedType);
		int[] res=new int[typeSize+1];
		int numComps=ExpressionTranslator.getTypeSize(containedType.type1);
		int[] gaps=determineGapsInSetBitVector(superType.type2, containedType.type2);
		int componentTypeSize=gaps.length-1;
		int[] accumulatedSizes=new int[numComps];
		accumulatedSizes[0]=1;
		int size2=ExpressionTranslator.getTypeSize(superType.type2);
		for(int i=1;i<numComps;i++)
		{
			accumulatedSizes[i]=accumulatedSizes[i-1]*size2;
		}			
		int lowerOffset=0;
		for(int j=0;j<numComps;j++)
		{
			lowerOffset+=accumulatedSizes[j]*gaps[0];
		}
		res[0]=lowerOffset;
		int[] subIndices=new int[numComps];
		
		for(int i=0;i<typeSize;i++)
		{
			int gapLength=0;
			if(i==typeSize-1)
			{
				for(int k=0;k<numComps;k++)
				{
					if(k<numComps-1)
					{
						gapLength+=gaps[subIndices[k]+1]*accumulatedSizes[k];
						gapLength+=gaps[subIndices[k+1]+1]*accumulatedSizes[k+1];
					}
				}
			}
			else
			{
				for(int k=0;k<numComps;k++)
				{
					if(subIndices[k]==componentTypeSize-1)
					{
						if(k<numComps-1)
						{
							gapLength+=gaps[subIndices[k]+1]*accumulatedSizes[k];
							gapLength+=gaps[0]*accumulatedSizes[k];
							gapLength+=gaps[subIndices[k+1]+1]*accumulatedSizes[k+1];
						}
						subIndices[k]=0;
					}
					else
					{
						if(k==0)
						{
							gapLength+=gaps[subIndices[0]+1];
						}
						subIndices[k]=subIndices[k]+1;
						break;
					}
				}
			}
			res[i+1]=gapLength;
		}
		return res;
	}
	
	/***************************************************************************
	 * For the property of the result see the description of determineGapsInSetBitVector.
	 * @param superType The contained type of interest.
	 * @param containedType The super type of interest.
	 * @return An array of length |containedType|+1
	 **************************************************************************/
	private static int[] determineGapsInSetBitVector_Sets(SetType superType,SetType containedType)
	{
		int typeSize=ExpressionTranslator.getTypeSizeConsiderBuffer(containedType, 1);
		int[] res=new int[typeSize+1];
		int[] gaps=determineGapsInSetBitVector(superType.type, containedType.type);
		int[] accumulatedGaps=new int[gaps.length];			
		accumulatedGaps[0]=gaps[0];
		for(int i=1;i<gaps.length;i++)
		{
			accumulatedGaps[i]=accumulatedGaps[i-1]+gaps[i];
		}
		res[0]=0;//the "first" element of set with respect to our enumeration is always the emptyset.
		int old_indexInSuperSet=0;
		for(int i=1;i<=typeSize;i++)
		{
			int k=i;
			int j=0;
			int indexInSuperSet=0;
			while(k!=0)
			{
				indexInSuperSet+=(k%2)*Utilities.pow2(j+accumulatedGaps[j]);
				k/=2;
				j++;
			}
			res[i]=indexInSuperSet-old_indexInSuperSet-1;
			old_indexInSuperSet=indexInSuperSet;
		}
		return res;
	}
	
	/***************************************************************************
	 * For the property of the result see the description of determineGapsInSetBitVector.
	 * @param superType The contained type of interest. The base type has to be a SetType.
	 * @param containedType The super type of interest.
	 * @return An array of length |containedType|+1
	 **************************************************************************/
	private static int[] determineGapsInSetBitVector_SubTypes(SubType subT)
	{
		Type baseT=Utilities.getBaseType(subT);
		int typeSize=ExpressionTranslator.getTypeSizeConsiderBuffer(subT, 1);
		int[] result=new int[typeSize+1];
		Seq<Value> subSeq=Values.getValueSeq(subT);
		Seq<Value> baseSeq=Values.getValueSeq(baseT);
		int i=0;
		int counter=0;
		while (true)
    {
			int gap=0;
      Next<Value> subNext = subSeq.get();
      if (subNext == null) break;
      Value subVal=subNext.head;
      subSeq = subNext.tail;
      Next<Value> baseNext = baseSeq.get();
      Value baseVal=baseNext.head;//if there is a value in the sequence of the subtype then there must be a value in this seq => no null check
      baseSeq = baseNext.tail;
      while(!subVal.equals(baseVal))
      {
      	gap++;
      	counter++;
      	baseNext = baseSeq.get();
      	baseVal=baseNext.head;
      	baseSeq = baseNext.tail;
      }
      result[i]=gap;
      i++;
      counter++;
    }
		result[typeSize]=ExpressionTranslator.getTypeSize(baseT)-counter;
		return result;
	}

	/***************************************************************************
	 * Generate an enumeration function for the given type and return the
	 * associated information.
	 * As every bit vector of the used sort represents a set value and because
	 * of the enumeration of set values we use here, we can use the identity function
	 * as an enumerating function here.
	 * @param sT The type for which an enumeration function shall be generated.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionNonSubType(SetType sT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		int len=ExpressionTranslator.getTypeSize(sT);
		int vectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		String variable="x";
		expTrans.addDeclaration("(define-fun "+funName+" (("+ variable +"(_ BitVec "+Integer.toString(vectorLength)+"))) (_ BitVec "+
				Integer.toString(vectorLength)+")"+variable+")");
		return new TypeEnumerationInformation(funName, vectorLength, len);
	}
	
	/***************************************************************************
	 * Generates an enumeration function for the values given by lst.
	 * The result of the enumeration is given by the position 
	 * of the respective value in the list.
	 * The enumeration function is defined via a table.
	 * @param lst The values used to generate the enumeration function.
	 * @param sT The elements of lst must correspond to the type sT. Must be canonical.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTable(List<Value> lst,SetType sT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		long len=lst.size();
		int typeVectorLength=ExpressionTranslator.getTypeSize(sT.type);
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		expTrans.addDeclaration("(declare-fun "+funName+" ((_ BitVec "+Integer.toString(typeVectorLength)+")) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+"))");
		long i=0;
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, sT);
		for(Value val:lst)
		{
			expTrans.addDeclaration("(assert (= ("+funName+" "+valC.process(val)+") "+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+"))");
			i++;
		}
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
	
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTableWithFunctionDefinition(List<Value> lst,SetType sT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		long len=lst.size();
		int typeVectorLength=ExpressionTranslator.getTypeSize(sT.type);
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		Iterator<Value> it=lst.iterator();
		StringBuilder strB=new StringBuilder();
		String variable="x";
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, sT);
		for(int i=0;i<len-1;i++)
		{
			Value val=it.next();
			strB.append("(ite (= "+variable+" "+valC.process(val)+")"+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+" ");
		}
		strB.append(IntTranslator.generateSMT_LIB_Constant(len-1, argumentVectorLength));
		strB.append(")".repeat((int) (len-1)));
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(typeVectorLength)+"))) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+") "+strB.toString()+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
	
	static class SetValueContainer extends ExpressionTranslator.ValueContainer
	{
		/**
		 * The length the translation of the values shall have.
		 */
		private int setVectorLength;
		/**
		 * The values of the SetType.
		 */
		private List<Value> elemValues;
		SetValueContainer(ExpressionTranslator expTrans,SetType sT)
		{
			setVectorLength=ExpressionTranslator.getTypeSize(sT.type);
			Seq<Value> seq=Values.getValueSeq(sT.type);
			elemValues=seq.list();
		}
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String process(Value val) 
		{
			return "#b"+processToBareVector(val);
		}
		/***************************************************************************
		 * Generate a bit sequence representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return A bit sequence representing val.
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			Set set=(Set)val;
			StringBuilder strB=new StringBuilder(setVectorLength);
			ListIterator<Value> it=elemValues.listIterator(elemValues.size());
			while(it.hasPrevious())
			{
				if(set.contains(it.previous()))
				{
					strB.append("1");
				}
				else
				{
					strB.append("0");
				}
			}
			return strB.toString();
		}
	}
	
	/***************************************************************************
	 * Translate a sequence of values with base type type to a list of SMT-LIB values.
	 * Add the #b prefix to these representations depending on bare.
	 * @param type The base type of the values. Must be canonical.
	 * @param seq A sequence of values whose basetype is type.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing the SMT-LIB representations of the
	 * values in seq.
	 **************************************************************************/
	List<String> processValueSequence(Type type,Seq<Value> seq,boolean bare)
	{
		ValueContainer valC=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans, type);
		int len=ExpressionTranslator.getTypeSize(type);
		List<String> result=new ArrayList<String>(len);
		Function <Value, String> proc;
		if(bare)
			proc=v -> valC.processToBareVector(v);
		else
			proc=v -> valC.process(v);
		while (true)
    {
      Next<Value> next = seq.get();
      if (next == null) break;
      result.add(proc.apply(next.head));
      seq = next.tail;
    }
		return result;
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Add the #b prefix to these representations depending on bare.
	 * @param sT The type whose values shall be computed. Has to be canonical.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing all values of the given type 
	 * represented by their corresponding SMT-LIB strings.
	 **************************************************************************/
	public static List<String> computeValues(SetType sT,boolean bare)
	{
		Type ty=sT.type;
		int len=ExpressionTranslator.getTypeSize(ty);
		int setLen=ExpressionTranslator.getTypeSize(sT);
		List<String> result=new ArrayList<String>(setLen);
		for(int i=0;i<setLen;i++)
		{
			if(bare)
			{
				result.add(IntTranslator.buildBitSequence(i, len));
			}
			else
			{
				result.add(IntTranslator.generateSMT_LIB_Constant(i, len));
			}		
		}
		return result;
	}
	
	/***************************************************************************
	 * Give a bit vector of length len containing only zeros.
	 * This bit vector represents the empty set for SetTypes with an
	 * underlying type of size len.
	 * @param len The length the bit vector shall have.
	 * @return A bit vector expression of length len.
	 **************************************************************************/
	private static String giveEmptySet(int len)
	{
		return "#b"+Integer.toString(0).repeat(len);
	}
	
	/***************************************************************************
	 * i has to be a bit vector expression representing a number x
	 * 0<=x<len. Generate a bit vector of length len
	 * that represents the set containing the x th element of the universe of the SetType
	 * @param i Determines the element of the singleton set.
	 * Has to be a bit vector expression of length len.
	 * @param len The length the bit vector shall have
	 * @return A bit vector expression of length len
	 * representing the ith element of the respective universe
	 **************************************************************************/
	private static String setIthElement(String i,int len)
	{
		return "(bvshl #b"+Integer.toString(0).repeat(len-1)+"1 "+i+")";
	}
	
	/***************************************************************************
	 * Generates a bit vector representing the singleton set containing the
	 * ith element of the respective universe of the settype.
	 * @param i Denotes the element of the singleton set. 0<=i<len
	 * @param len The length the bit vector shall have
	 * @return A bit vector expression of length len
	 * representing the ith element of the respective universe
	 **************************************************************************/
	private static String setIthElement(int i,int len)
	{
		return "#b"+"0".repeat(len-1-i)+"1"+"0".repeat(i);
	}

	/***************************************************************************
	 * Generates a function that takes the bit vector representation for
	 * a number x 0<=x<=|s.type| and returns a bit vector of length |s|
	 * that represents the set that contains all the sets of cardinality x
	 * @param s The type of interest
	 * @return A bit vector expression representing a set of sets
	 **************************************************************************/
	private String generateFiltersForSetsOfCardN(SetType s)
	{
		String res=filtersForElementsOfCardinalityN.get(s);
		if(res!=null)
			return res;
		Type elemType=s.type;
		int setlen=ExpressionTranslator.getTypeSize(s);
		int len=ExpressionTranslator.getTypeSize(elemType);
		int elemlength=IntTranslator.determineBitVectorLength(0, len);
		res=expTrans.getFunctionName(defaultCardinalityFilterName);
		if(expTrans.getUnquantify())
		{
			expTrans.addDeclaration("(declare-fun "+res+" ((_ BitVec "+Integer.toString(elemlength)+"))(_ BitVec "+Integer.toString(setlen)+"))");
			expTrans.addDeclaration("(assert (= ("+res+" "+IntTranslator.generateSMT_LIB_Constant(0, elemlength)+")"+setIthElement(0, setlen)+"))");
			for(int i=1;i<=len;i++)
			{
				StringBuilder strB=new StringBuilder();
				for(int j=setlen-1;j>=0;j--)
				{
					if(Integer.bitCount(j)==i)
					{
						strB.append("1");
					}
					else
					{
						strB.append("0");
					}
				}
				String arg=IntTranslator.generateSMT_LIB_Constant(i, elemlength);
				expTrans.addDeclaration("(assert (= ("+res+" "+arg+")"+"#b"+strB.toString()+"))");
			}
		}
		else
		{
			String variable="x";
			StringBuilder strB=new StringBuilder();
			strB.append("(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(0, elemlength)+")"+
					setIthElement(0, setlen)+" ");
			for(int i=1;i<=len;i++)
			{
				StringBuilder builderForSet=new StringBuilder();
				for(int j=setlen-1;j>=0;j--)
				{
					if(Integer.bitCount(j)==i)
					{
						builderForSet.append("1");
					}
					else
					{
						builderForSet.append("0");
					}
				}
				if(i<len)
				{
					strB.append("(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(i, elemlength)+")"+
							"#b"+builderForSet.toString()+" ");
				}
				else
				{
					strB.append("#b"+builderForSet.toString());
				}			
			}
			strB.append(")".repeat(len));
			expTrans.addDeclaration("(define-fun "+res+" ((x (_ BitVec "+Integer.toString(elemlength)+")))"
					+ "(_ BitVec "+Integer.toString(setlen)+") "+strB.toString()+")");
		}
		filtersForElementsOfCardinalityN.put(s, res);
		return res;
	}
	
	/***************************************************************************
	 * Generate the bit vector expression that represents the set
	 * of all sets containing the respective elem^th element.
	 * @param len The length of the resulting bit vector len>0
	 * @param elem The index of the respective value. elem>=0
	 * @return A bit vector expression of length len
	 **************************************************************************/
	private static String generateElementFilter(int len,int elem)
	{
		int aux1=Utilities.pow2(elem);
		int aux2=len/(2*aux1);
		return "#b"+("1".repeat(aux1)+"0".repeat(aux1)).repeat(aux2);
	}	
	
	public String visit(SubsetExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(= (bvand "+str1+" "+str2+") "+str1+")";
	}
	
	public String visit(InSetExp exp)
	{
		String str1=expTrans.compute(exp.exp1);
		String str2=exp.exp2.accept(this);
		TypeEnumerationInformation info=expTrans.getTypeEnumerationFunction(usedType.type);
		int diff=setVectorLength-(int)info.enumerationFunctionResultVectorLength;
		str1=expTrans.adaptTranslation(str1, exp.exp1.getType(), usedType.type);
		String aux="("+info.functionName+" "+str1+")";
		if(diff!=0)
		{
			aux="((_ zero_extend "+Integer.toString(diff)+")"+aux+")";
		}
		aux=setIthElement(aux, setVectorLength);
		return "(distinct (bvand "+aux+" "+str2+")"+giveEmptySet(setVectorLength)+")";
	}
	
	public String visit(SetSizeExp exp)
	{
		Integer j=Integer.valueOf(setVectorLength);
		String cardName=cardinalityFunctionNames.get(j);
		if(cardName==null)
		{
			int resultLength=IntTranslator.determineBitVectorLength(0, setVectorLength);
			cardName=expTrans.getFunctionName(defaultCardinalityName);
			cardinalityFunctionNames.put(j, cardName);
			String variable="x";
			String dec;
			if(setVectorLength==1)//boolector does not support (bvadd a)
			{
				dec="(define-fun "+cardName+" (("+variable+" (_ BitVec "+Integer.toString(setVectorLength)+")))"
						+ "(_ BitVec "+Integer.toString(resultLength)+") (ite (= "+variable+" #b1) #b1 #b0))";
			}
			else
			{
				String one=IntTranslator.generateSMT_LIB_Constant(1, resultLength);
				String zero=IntTranslator.generateSMT_LIB_Constant(0, resultLength);
				StringBuilder strB=new StringBuilder();
				for(int i=0;i<setVectorLength;i++)
				{
					strB.append("(ite (= "+setIthElement(i, setVectorLength)+" (bvand "+variable+" "+setIthElement(i, setVectorLength)+"))"+one+" "+zero+")");
				}
				dec="(bvadd "+strB.toString()+")";
				dec="(define-fun "+cardName+" (("+variable+" (_ BitVec "+Integer.toString(setVectorLength)+")))"
						+ "(_ BitVec "+Integer.toString(resultLength)+") "+dec+" )";
			}		
			expTrans.addDeclaration(dec);
		}
		return "("+cardName+" "+exp.exp.accept(this)+")";
	}
	
	public String visit(EqualsExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(= "+str1+" "+str2+")";
	}
	
	public String visit(NotEqualsExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(distinct "+str1+" "+str2+")";
	}
	
	public String visit(BigUnionExp exp)
	{
		String baseSet=compute(exp.exp);
		int len1=getLength(exp.exp);
		int len2=getLength(exp);
		String funName=bigUnion.get(len1);
		if(funName==null)
		{
			funName=expTrans.getFunctionName(defaultBigUnionName);
			String variable="x";
			StringJoiner strJ=new StringJoiner(" ");
			for(int i=0;i<len2;i++)
			{
				String filter=generateElementFilter(len1, i);
				strJ.add("(ite (distinct (bvand "+filter+" "+variable+") "+giveEmptySet(len1)+")"+setIthElement(i, len2)+" "+giveEmptySet(len2)+")");
			}
			String dec="(define-fun "+funName+" (("+variable+" (_ BitVec "+Integer.toString(len1)+")))"
					+ "(_ BitVec "+Integer.toString(len2)+") (bvor "+strJ.toString()+"))";
			expTrans.addDeclaration(dec);
			bigUnion.put(len1, funName);
		}
		String res="("+funName+" "+baseSet+")";
		return enlargeSet(res, exp.getType(), usedType);
	}
	
	public String visit(BigIntersectExp exp)
	{
		String baseSet=compute(exp.exp);
		int len1=getLength(exp.exp);
		int len2=getLength(exp);
		String funName=bigIntersection.get(len2);
		if(funName==null)
		{
			funName=expTrans.getFunctionName(defaultBigIntersectionName);
			String variable="x";
			StringJoiner strJ=new StringJoiner(" ");
			for(int i=0;i<len2;i++)
			{
				String filter=generateElementFilter(len1, i);
				strJ.add("(ite (= (bvand "+filter+" "+variable+") "+variable+")"+setIthElement(i, len2)+" "+giveEmptySet(len2)+")");
			}
			String dec="(define-fun "+funName+" (("+variable+" (_ BitVec "+Integer.toString(len1)+")))"
					+ "(_ BitVec "+Integer.toString(len2)+") (bvor "+strJ.toString()+"))";
			expTrans.addDeclaration(dec);
			bigIntersection.put(len2, funName);
		}
		String res="("+funName+" "+baseSet+")";
		return enlargeSet(res, exp.getType(), usedType);
	}
	
	
	public String visit(UnionExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvor "+str1+" "+str2+")";
	}
	
	public String visit(IntersectExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvand "+str1+" "+str2+")";
	}
	
	public String visit(WithoutExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvand "+str1+" (bvnot "+str2+"))";
	}
	
	/***************************************************************************
	 * Create the SMT-LIB representation of a powerset.
	 * @param set The SMT-LIB representation of the set that shall be used for the
	 * construction of the power set.
	 * @param powSetLength The length of the bit vector representing the power set.
	 * @param setLength The length of set.
	 * @return A SMT-LIB representation for the power set of set.
	 **************************************************************************/
	private String powerSetAux(String set,int powSetLength,int setLength)
	{
		String powerSetFunName=powerSetFunctions.get(powSetLength);
		if(powerSetFunName==null)
		{
			powerSetFunName=expTrans.getFunctionName(defaultPowerSetFunctionName);
			String variable="x";
			String dec="(define-fun "+powerSetFunName+" (("+variable+" (_ BitVec "+Integer.toString(setLength)+")))"
					+ "(_ BitVec "+Integer.toString(powSetLength)+")";
			StringJoiner strJ=new StringJoiner(" ");
			for(int i=0;i<setLength;i++)
			{
				String s=setIthElement(i, setLength);
				String filter=generateElementFilter(powSetLength,i);
				strJ.add("(ite (= "+giveEmptySet(setLength)+" (bvand "+s+" "+variable+"))"+filter+" "+giveEmptySet(powSetLength)+")");
			}
			String aux;
			if(setLength==1)
			{
				aux=strJ.toString();
			}
			else
			{
				aux="(bvor "+strJ.toString()+")";
			}
			dec+="(bvnot "+aux+"))";
			expTrans.addDeclaration(dec);
			powerSetFunctions.put(powSetLength, powerSetFunName);
		}
		return "("+powerSetFunName+" "+set+")";
	}
	
	public String visit(PowersetExp exp)
	{
		String set=compute(exp.exp);
		int len1=getLength(exp);
		int len2=getLength(exp.exp);
		String res=powerSetAux(set, len1,len2);
		return enlargeSet(res, exp.getType(), usedType);		
	}
	
	public String visit(Powerset1Exp exp)
	{
		SetType s=((SetType)exp.exp1.getType());
		String set=compute(exp.exp1);
		int len1=getLength(exp);
		int len2=getLength(exp.exp1);
		String res=powerSetAux(set, len1,len2);
		String filter=generateFiltersForSetsOfCardN(s);
		IntType iT=new IntType(0, len2);
		iT.setCanonical();
		String card=expTrans.compute(exp.exp2,iT);
		res="(bvand "+res+" ("+filter+" "+card+"))";
		return enlargeSet(res, exp.getType(), usedType);		
	}
	
	public String visit(Powerset2Exp exp)
	{
		SetType s=((SetType)exp.exp1.getType());
		String set=compute(exp.exp1);
		int len1=getLength(exp);
		int len2=getLength(exp.exp1);
		String res=powerSetAux(set, len1,len2);
		String filter=generateFiltersForSetsOfCardN(s);
		IntType iT1=(IntType)Utilities.getBaseType(exp.exp2.getType());
		IntType iT2=(IntType)Utilities.getBaseType(exp.exp3.getType());
		int lower=iT1.ivalue1.intValue();
		int upper=iT2.ivalue2.intValue();
		IntType iT=new IntType(lower, upper);
		iT.setCanonical();
		int len4=IntTranslator.determineBitVectorLength(lower, upper);
		assert(lower<=upper);
		String cardLowerBound=expTrans.compute(exp.exp2, iT);
		String cardUpperBound=expTrans.compute(exp.exp3, iT);
		StringJoiner strJ=new StringJoiner(" ");
		for(int i=lower;i<=upper;i++)
		{
			String val=IntTranslator.generateSMT_LIB_Constant(i, len4);
			strJ.add("(ite (and (bvule "+cardLowerBound+" "+val+")(bvule "+val+" "+cardUpperBound+"))("+filter+" "+val+")"+giveEmptySet(len1)+")");
		}
		String aux;
		if(lower<upper)
		{
			aux="(bvor "+strJ.toString()+")";
		}
		else
		{
			aux=strJ.toString();
		}
		res="(bvand "+res+" "+aux+")";
		return enlargeSet(res, exp.getType(), usedType);		
	}
	
	private String generateCartesian(Expression exp1,String set,int setLength)
	{
		String set1=expTrans.compute(exp1);
		int len1=getLength(exp1);
		String filter=setIthElement(0, setLength);
		String aux="(= "+filter+" (bvand "+filter+" "+set+"))";
		String res="(ite "+aux+" "+set1+" "+giveEmptySet(len1)+")";
		if(setLength!=1)
		{
			for(int i=1;i<setLength;i++)
			{
				filter=setIthElement(i, setLength);
				aux="(= "+filter+" (bvand "+filter+" "+set+"))";
				aux="(ite "+aux+" "+set1+" "+giveEmptySet(len1)+")";
				res="(concat "+aux+" "+res+")";				
			}
		}
		return res;
	}
	
	public String visit(CartesianExp exp)
	{
		int pos=exp.exps.length;
		String set=expTrans.compute(exp.exps[pos-1]);
		int len=getLength(exp.exps[pos-1]);
		String res=generateCartesian(exp.exps[pos-2], set,len);
		len=len*getLength(exp.exps[pos-2]);
		for(int i=exp.exps.length-3;i>=0;i++)
		{
			res=generateCartesian(exp.exps[i], res,len);
		}
		return res;	
	}
	
	public String visit(EmptySetExp exp)
	{
		return giveEmptySet(setVectorLength);
	}
	
	public String visit(SetBuilderExp exp)
	{
		String expStr=expTrans.compute(exp.exp);
		QuantifiedVariable qv=exp.qvar;
		QuantifiedVariableCore[] qvc=qv.qvcore;
		TypeEnumerationInformation info=expTrans.getTypeEnumerationFunction(exp.exp.getType());
		int diff=setVectorLength-(int)info.enumerationFunctionResultVectorLength;
		java.util.Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp.exp);
		for(int i=qvc.length-1;i>=0;i--)
		{
			IdentifierTypeQuantifiedVar itqv=(IdentifierTypeQuantifiedVar)qvc[i];
			String variable=itqv.ident.string;
			String condition=null;
			if(itqv.wexp!=null)
			{
				condition=expTrans.compute(itqv.wexp);
				freeSyms.addAll(expTrans.typedFreeSymbolsWithoutValues(itqv.wexp));
			}
			Type t=((ValueSymbol)itqv.ident.getSymbol()).type;
			if(t instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
			{
				String cond=expTrans.generateTypeConditions(t, variable);
				condition=condition==null?cond:"(and "+condition+" "+cond+")";
			}
			if(i==qvc.length-1)
			{
				expStr="((_ zero_extend "+Integer.toString(diff)+")("+info.functionName+" "+expStr+"))";
				expStr=setIthElement(expStr, setVectorLength);
			}			
			if(condition!=null)
			{
				expStr="(ite "+condition+" "+expStr+" "+giveEmptySet(setVectorLength)+")";
			}
			List<String> vals=expTrans.computeValues(t);
			if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), expStr.length()))
			{
				String typeString="(_ BitVec "+Integer.toString(setVectorLength)+")";
				if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
				{
					List<String> lst=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms, (ValueSymbol)itqv.ident.getSymbol(), typeString);
					String args=String.join(" ", lst);
					if(lst.size()==1)
					{
						expStr=args;
					}
					else
					{
						expStr="(bvor "+args+")";
					}
					freeSyms.remove(itqv.ident.getSymbol());
					continue;
				}
				else
				{
					expStr=expTrans.generateAuxillaryExpansionFunction(expStr, freeSyms,typeString);					
				}				
			}
			String aux=expStr;
			expStr=vals.stream().map(v -> "(let (("+variable+" "+v+")) "+aux+")").collect(Collectors.joining(" "));
			if(vals.size()!=1)
			{
				expStr="(bvor "+expStr+")";
			}
			freeSyms.remove(itqv.ident.getSymbol());
		}
		return expStr;
	}
	
	public String visit(IntervalExp exp)
	{
		Expression e1=exp.exp1;
		Expression e2=exp.exp2;
		IntType iT1=(IntType)Utilities.getBaseType(e1.getType());
		IntType iT2=(IntType)Utilities.getBaseType(e2.getType());
		if(iT1.ivalue1>iT2.ivalue2)
			return giveEmptySet(setVectorLength);
		IntType iT3=new IntType(iT1.ivalue1, iT2.ivalue2);
		iT3.setCanonical();
		if(iT1.ivalue1.equals(iT2.ivalue2))
		{
			Expression aux=new EqualsExp(e1, e2);
			String set="(ite "+expTrans.compute(aux)+" #b1 #b0)";
			return enlargeSet(set, new SetType(iT3), usedType);
		}
		IntType iT4=new IntType(Math.min(iT1.ivalue1, iT2.ivalue1), Math.max(iT1.ivalue2, iT2.ivalue2));
		iT4.setCanonical();
		Seq<Value> seq=Values.getValueSeq(iT3);
		List<String> vals=IntTranslator.processValueSequence(iT4, seq,false);
		Iterator<String>it=vals.iterator();
		String s1=expTrans.compute(e1, iT4);
		String s2=expTrans.compute(e2, iT4);
		int len=ExpressionTranslator.getTypeSize(iT3);
		String emptyset=giveEmptySet(len);
		StringBuilder strB=new StringBuilder();
		strB.append("(bvor");
		int count=0;
		if(iT4.ivalue1<0)
		{			
			while(it.hasNext())
			{
				String val=it.next();
				String setVal=setIthElement(count, len);
				strB.append("(ite (and (bvsle "+s1+" "+val+") (bvsle "+val+" "+s2+"))"+setVal+" "+emptyset+")");
				count++;
			}						
		}
		else
		{
			while(it.hasNext())
			{
				String val=it.next();
				String setVal=setIthElement(count, len);
				strB.append("(ite (and (bvule "+s1+" "+val+") (bvule "+val+" "+s2+"))"+setVal+" "+emptyset+")");
				count++;
			}
		}
		strB.append(")");
		return enlargeSet(strB.toString(), new SetType(iT3), usedType);
	}
	
	public String visit(EnumeratedSetExp exp)
	{
		TypeEnumerationInformation info=expTrans.getTypeEnumerationFunction(usedType.type);
		String enumFun=info.functionName;
		int diff=setVectorLength-(int)info.enumerationFunctionResultVectorLength;
		if(exp.exps.length==1)//in boolector bit wise or needs at least 2 arguments.
		{
			if(diff==0)
			{
				return setIthElement("("+enumFun+" "+expTrans.compute(exp.exps[0], usedType.type)+")", setVectorLength);
			}
			else
			{
				String aux="((_ zero_extend "+Integer.toString(diff)+")("+enumFun+" "+expTrans.compute(exp.exps[0], usedType.type)+"))";
				return setIthElement(aux, setVectorLength);
			}
		}
		else
		{
			String elems;
			if(diff==0)
			{
				elems=Arrays.stream(exp.exps).map(e -> 
				setIthElement("("+enumFun+" "+expTrans.compute(e, usedType.type)+")",setVectorLength))
						.collect(Collectors.joining(" "));
			}
			else
			{
				elems=Arrays.stream(exp.exps).map(e -> 
				setIthElement("((_ zero_extend "+Integer.toString(diff)+")("+enumFun+" "+expTrans.compute(e, usedType.type)+"))",setVectorLength))
						.collect(Collectors.joining(" "));
			}
			return "(bvor "+elems+")";
		}		
	}
	
	public String visit(IdentifierExp exp)
	{
		String set=expTrans.compute(exp);
		set=enlargeSet(set, exp.getType(), usedType);
		return set;
	}
	
	public String visit(ApplicationExp exp)
	{
		String set=expTrans.compute(exp);
		set=enlargeSet(set, exp.getType(), usedType);
		return set;
	}
	
	public String visit(Binder b)
	{
		String str=expTrans.compute(b.exp);
		return "( "+b.ident.string+" "+str+")";
	}
	
	public String visit(LetParExp exp)
	{
		String str=Arrays.stream(exp.binder).map(b -> b.accept(this)).collect(Collectors.joining(" "));
		String expStr=exp.exp.accept(this);
		return "(let ("+str+") "+expStr+")";
	}
	
	public String visit(IfThenElseExp exp)
	{
		String cond=expTrans.compute(exp.exp1);
		String thenExp=exp.exp2.accept(this);
		String elseExp=exp.exp3.accept(this);
		return "(ite "+cond+" "+thenExp+" "+elseExp+")";
	}
	
	public String visit(TupleSelectionExp exp)
	{
		String set=expTrans.compute(exp);
		set=enlargeSet(set, exp.getType(), usedType);
		return set;
	}
	
	public String visit(MapSelectionExp exp)
	{
		String set=expTrans.compute(exp);
		set=enlargeSet(set, exp.getType(), usedType);
		return set;
	}
	
	public String visit(SetType sT)
	{
		int len=ExpressionTranslator.getTypeSize(sT.type);
		return "(_ BitVec "+Integer.toString(len)+")";
	}


	/***************************************************************************
	 * Shall not be used - Not sufficiently tested + inefficient
	 * Usually each value of the bit vector sort that corresponds to sT
	 * represents a value of setStr.
	 * This does not apply only if a SubType occurs in the universe of sT
	 * and SMT.expandBaseTypesOfSubTypes is true.
	 * If SMT.expandBaseTypesOfSubTypes=true then the sets are represented with respect
	 * to the base type. 
	 * In this case this method generates a boolean valued SMT-LIB expression
	 * that is true iff setStr represents a value of sT.
	 **************************************************************************/
	public String generateTypeConditions(SetType sT,String setStr)
	{
		if(!SMT.expandBaseTypesOfSubTypes)
			return null;
		if(!Utilities.containsSubType(sT))
			return null;
		Type elemType=sT.type;
		List<String> vals=expTrans.computeValues(elemType);
		int counter=0;
		int len=vals.size();
		List<String> lst=new ArrayList<String>();
		for(String v:vals)
		{
			String cond=expTrans.generateTypeConditions(elemType, v);
			if(cond==null)
				return null;
			lst.add("(ite "+cond+" "+"#b"+"0".repeat(len-1-counter)+"1"+"0".repeat(counter)+" "+"0".repeat(len)+")");
		}
		String aux;
		if(lst.size()==1)
		{
			aux=String.join(" ", lst);
		}
		else
		{
			aux="(bvor "+String.join(" ", lst)+")";
		}		
		return "(= "+setStr+" (bvand "+setStr+" "+aux+"))";
	}

	public long giveVectorLength(SetType type)
	{
		return ExpressionTranslator.getTypeSize(type.type);
	}
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
