// ---------------------------------------------------------------------------
// Utilities.java
// This class provides utility methods and classes used in the SMT package.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: Utilities.java,v 1.4 2020/06/09 13:02:23 schreine Exp $
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.stream.Collectors;

import riscal.syntax.AST;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.AndExp;
import riscal.syntax.AST.Expression.ExistsExp;
import riscal.syntax.AST.Expression.ForallExp;
import riscal.syntax.AST.Expression.ImpliesExp;
import riscal.syntax.AST.Expression.TrueExp;
import riscal.syntax.AST.Identifier;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IdentifierType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.MapType;
import riscal.syntax.AST.Type.RecordType;
import riscal.syntax.AST.Type.RecursiveType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.AST.Type.UnitType;
import riscal.types.Environment.Symbol.ValueSymbol;

class Utilities 
{
	/***************************************************************************
	 * A pair of values of the given types.
	 **************************************************************************/
	public static class Pair<S,T>
	{
		public S first;
		public T second;
		public Pair(S s, T t)
		{
			this.first=s;
			this.second=t;
		}
	}
	
	/***************************************************************************
	 * A triple of values of the given types
	 **************************************************************************/
	public static class Triple<T1,T2,T3>
	{
		public T1 first;
		public T2 second;
		public T3 third;
		public Triple(T1 x,T2 y,T3 z)
		{
			first=x;
			second=y;
			third=z;
		}
	}
	
	public static class Conditions
	{
		private Set<ValueSymbol> values;
		private Stack<Set<ValueSymbol>> oldVariablesWithConditions=new Stack<Set<ValueSymbol>>();
		private Stack<Map<ValueSymbol,List<Expression>>> oldSimpleConditions=new Stack<Map<ValueSymbol,List<Expression>>>();
		private Set<ValueSymbol> variablesWithConditions=new HashSet<ValueSymbol>();
		private Map<ValueSymbol,List<Expression>> simpleConditions=new HashMap<ValueSymbol, List<Expression>>();
		
		public Conditions(Set<ValueSymbol> values)
		{
			this.values=values;
		}
		
		public void clearAll()
		{
			oldVariablesWithConditions.clear();
			oldSimpleConditions.clear();
			variablesWithConditions.clear();
			simpleConditions.clear();
		}
		
		public List<Expression> getSimpleCondition(ValueSymbol sym)
		{
			return simpleConditions.get(sym);
		}
		
		public boolean hashCondition(ValueSymbol sym)
		{
			return variablesWithConditions.contains(sym);
		}
		
		public void revertVariablesWithConditions()
		{
			variablesWithConditions=oldVariablesWithConditions.pop();
		}
		
		public void revertSimpleConditions()
		{
			simpleConditions=oldSimpleConditions.pop();
			variablesWithConditions=oldVariablesWithConditions.pop();
		}
		
		private Map<ValueSymbol,List<Expression>> copyConditionMap(Map<ValueSymbol,List<Expression>> x)
		{
			Map<ValueSymbol,List<Expression>> m=new HashMap<ValueSymbol, List<Expression>>();
			for(Map.Entry<ValueSymbol,List<Expression>> e:x.entrySet())
			{
				List<Expression> lst=new ArrayList<AST.Expression>(e.getValue());
				m.put(e.getKey(), lst);
			}
			return m;
		}
		
		public void analyseWithSimpleConditions(Expression exp)
		{
			Set<ValueSymbol> old1=new HashSet<ValueSymbol>(variablesWithConditions);
			oldVariablesWithConditions.push(old1);
//			Map<ValueSymbol,List<Expression>> olds=new HashMap<ValueSymbol, List<Expression>>(simpleConditions);
			Map<ValueSymbol,List<Expression>> olds=copyConditionMap(simpleConditions);
			oldSimpleConditions.push(olds);		
			List<Expression> conds=breakupFormula(exp);
			for(Expression e:conds)
			{
				Set<ValueSymbol> free=typedFreeSymbolsWithoutValues(e);
				if(free.size()==1)
				{
					ValueSymbol sym=free.iterator().next();
					List<Expression> lst=simpleConditions.get(sym);
					if(lst==null)
					{
						lst=new ArrayList<AST.Expression>();
						lst.add(e);
						simpleConditions.put(sym, lst);
					}
					else
					{
						lst.add(e);
					}
				}
				else
					variablesWithConditions.addAll(free);
			}		
		}
		
		public void analyse(Expression exp)
		{
			Set<ValueSymbol> old1=new HashSet<ValueSymbol>(variablesWithConditions);
			oldVariablesWithConditions.push(old1);
			Set<ValueSymbol> free=typedFreeSymbolsWithoutValues(exp);
			variablesWithConditions.addAll(free);
		}		
		
		public void analyse(Binder[] binders)
		{
			Set<ValueSymbol> old1=new HashSet<ValueSymbol>(variablesWithConditions);
			oldVariablesWithConditions.push(old1);
			for(Binder b:binders)
			{
				Expression exp=b.exp;
				Set<ValueSymbol> free=typedFreeSymbolsWithoutValues(exp);
				for(ValueSymbol vsym:free)
				{
					if(variablesWithConditions.contains(vsym))
					{
						variablesWithConditions.add((ValueSymbol) b.ident.getSymbol());
						break;
					}
					else if(simpleConditions.get(vsym)!=null)
					{
						variablesWithConditions.add((ValueSymbol) b.ident.getSymbol());
						break;
					}
				}
			}
		}
		
		public void analyse(QuantifiedVariable qv)
		{
			Set<ValueSymbol> old1=new HashSet<ValueSymbol>(variablesWithConditions);
			oldVariablesWithConditions.push(old1);
			Set<ValueSymbol> vars=new HashSet<ValueSymbol>();
			for(QuantifiedVariableCore qvc:qv.qvcore)
			{
				vars.add((ValueSymbol) qvc.ident.getSymbol());
				if(qvc.wexp!=null)
				{
					variablesWithConditions.add((ValueSymbol) qvc.ident.getSymbol());
				}
			}
			if(qv.exp!=null)
			{
				Set<ValueSymbol> free=typedFreeSymbolsWithoutValues(qv.exp);
				free.retainAll(vars);
				variablesWithConditions.addAll(free);
			}
		}
		
		/***************************************************************************
		 * Compute the ValueSymbol that correspond to unbound identifier in an expression.
		 * Remove the ValueSymbol that correspond to values introduced by the current
		 * declarations.
		 * @param exp The expression that shall be analysed.
		 * @return The ValueSymbol of the free variables.
		 **************************************************************************/
		private Set<ValueSymbol> typedFreeSymbolsWithoutValues(Expression exp)
		{
			Set<ValueSymbol> freeVariables=TypedFreeVariables.compute(exp);
			freeVariables.removeAll(values);
			return freeVariables;
		}
		
		private static List<Expression> breakupFormula(Expression exp)
		{
			if(exp instanceof AndExp)
			{
				AndExp a=(AndExp)exp;
				List<Expression> exps=breakupFormula(a.exp1);
				exps.addAll(breakupFormula(a.exp2));
				return exps;
			}
			List<Expression> exps=new ArrayList<AST.Expression>();
			exps.add(exp);
			return exps;
		}
	}
	
	/***************************************************************************
	 * Computes 2^x.
	 * @param x x>=0
	 * @return 2^x
	 **************************************************************************/
	public static int pow2(int x)
	{
		return 1<<x;
	}
	
  /***************************************************************************
   * Process and evaluate expression to an integer. 
   * Based on the tryEvalInt method from the TypeChecker class.
   * @param exp the expression
   * @return the integer value of the expression (null, if none)
   **************************************************************************/
	public static Long getIntValueFromExpression(Expression exp)
	{
    Type type=exp.getType();
    if (!(type instanceof IntType))
    	return null;
    IntType type0 = (IntType)type;
    
    // the type denotes its value
    if (type0.ivalue1.equals(type0.ivalue2)) return type0.ivalue1;
    
    return null;
	}
	
	/***************************************************************************
	 * Generates a string which is not contained in the set ids and adds it to ids.
	 * @param suggestion This string is used as a basis for the construction of the resulting string.
	 * @param ids This set gives the values the resulting string must not attain.
	 * 		The resulting string is added to this set.
	 * @return A string which was not contained in ids before the method invocation.
	 **************************************************************************/
		public static String NewIdentString(String suggestion, Set<String> ids)
		{
			String str=suggestion;
			int i=1;
			while(ids.contains(str))
			{
				str=suggestion+"_"+Integer.toString(i);
				i++;
			}
			ids.add(str);
			return str;
		}
		
		/***************************************************************************
		 * Generates a string which is neither in the set ids nor in the set constraints.
		 * The resulting string is added to ids.
		 * @param suggestion his string is used as a basis for the construction of the resulting string.
		 * @param ids This set gives values the resulting string must not attain.
		 * 	The resulting string is added to this set.
		 * @param constraints This set gives values the resulting string must not attain.
		 * @return A string which was not contained in ids before the method invocation.
		 * 	and which is not contained in constraints
		 **************************************************************************/
		public static String NewIdentStringWithConstraints(String suggestion, Set<String> ids,Set<String> constraints)
		{
			String str=suggestion;
			int i=1;
			while(ids.contains(str)||constraints.contains(str))
			{
				str=suggestion+"_"+Integer.toString(i);
				i++;
			}
			ids.add(str);
			return str;
		}
		
		/***************************************************************************
		 * Generates a string which is not contained in the set ids.
		 * @param suggestion This string is used as a basis for the construction of the resulting string.
		 * @param ids This set gives the values the resulting string must not attain.
		 * @return A string which was is not contained in ids.
		 **************************************************************************/
		public static String NewIdentStringWithoutSetUpdate(String suggestion, Set<String> ids)
		{
			String str=suggestion;
			int i=1;
			while(ids.contains(str))
			{
				str=suggestion+"_"+Integer.toString(i);
				i++;
			}
			return str;
		}
	
	
	/***************************************************************************
	 * Generates an array of strings whose elements are not part of ids and adds the elements to ids.
	 * @param suggestions These strings are used as a basis for the construction of the resulting strings.
	 * @param ids This set gives the values the resulting strings must not attain.
	 * 		The resulting strings are added to this set.
	 * @return Strings which were not contained in ids before the method invocation
	 * 		and which are mutually different.
	 **************************************************************************/
	public static String[] multipleNewIdentStrings(String[] suggestions,Set<String> ids)
	{
		int i;
		String str;
		String[] res =new String[suggestions.length];
		for(String s:suggestions)
		{
			i=1;
			str=s;
			while(ids.contains(str))
			{
				str=s+"_"+Integer.toString(i);
				i++;
			}
			ids.add(str);
			res[i]=str;
		}
		return res;
	}
	
	/***************************************************************************
	 * Determines the base type of a given type.
	 * @param type The type whose base type we want to determine.
	 * @return If type is not a subtype type otherwise getBaseType of the 
	 * 	base type of type.
	 **************************************************************************/
  public static Type getBaseType(Type type)
  {
    while (type instanceof SubType)
    {
      SubType type0 = (SubType)type;
      type = type0.type;
    }
    return type;
  }
  
  /***************************************************************************
   * Determines if a type contains a subtype.
   * @param type The type we want to check.
   * @return true if type is a subtype or any component type (e.g. domain and image type of a maptype)
   * 		contains a subtype.
   **************************************************************************/
  public static boolean containsSubType(Type type)
  {
  	if(type instanceof SubType)
  		return true;
  	if(type instanceof IntType)
  		return false;
  	if(type instanceof BoolType)
  		return false;
  	if(type instanceof UnitType)
  		return false;
  	if(type instanceof MapType)
  	{
  		MapType mT=(MapType)type;
  		if(containsSubType(mT.type2))
  			return true;
  		else
  			return false;
  	}
  	if(type instanceof TupleType)
  	{
  		TupleType tT=(TupleType)type;
  		for(Type t:tT.types)
  		{
  			if(containsSubType(t))
  				return true;
  		}
  		return false;
  	}
  	if(type instanceof SetType)
  	{
  		SetType sT=(SetType)type;
  		if(containsSubType(sT.type))
  			return true;
  		else
  			return false;
  	}
  	assert(false);
  	return false;
  }
	
  
  /***************************************************************************
   * Constructs an implication expression.
   * @param exp1 Shall not be null
   * @param exp2 Shall not be null.
   * @return If exp1 is the TrueExp exp2 otherwise an ImpliesExp where 
   * 		the premise is exp1 and the conlusion exp2.
   **************************************************************************/
	public static Expression constructImplicationExpression(Expression exp1,Expression exp2)
	{
//		if(exp1==null||exp1 instanceof TrueExp)
		if(exp1 instanceof TrueExp)
			return exp2;
		else
			return new ImpliesExp(exp1,exp2);		
	}
	
	/***************************************************************************
	 * Generates a non canonic type
	 * @param type The type we want to translate to a non canonic type.
	 * @return
	 * 		If type is not canonic or it is the BoolType or the UnitType then type.
	 * 		If it is an IntType an InType where we repalce the integer bounds with expressions.
	 * 		For the other types a type where we replace the contained types with the respective
	 * 			syntactic ones.
	 **************************************************************************/
	public static Type canonic2syntacticType(Type type)
	{
		if(!type.isCanonical()||type instanceof UnitType||type instanceof BoolType)
			return type;
		if(type instanceof riscal.syntax.AST.Type.IntType)
		{
			IntType iT=(IntType)type;
			return new IntType(AST.number(iT.ivalue1),AST.number(iT.ivalue2));
		}
		if(type instanceof MapType)
		{
			MapType mT=(MapType)type;
			Type type1=canonic2syntacticType(mT.type1);
			Type type2=canonic2syntacticType(mT.type2);
			return new MapType(type1,type2);
		}
		if(type instanceof TupleType)
		{
			TupleType tT=(TupleType)type;
			List<Type> types=Arrays.stream(tT.types).map(t -> canonic2syntacticType(t)).collect(Collectors.toList());
			return new TupleType(types);
		}
		if(type instanceof RecordType)//We immediately transform RecordTypes to TupleTypes
		{
			RecordType rT=(RecordType)type;
			ValueSymbol syms[]=rT.symbols;
			List<Type> types=Arrays.stream(syms).map(s -> canonic2syntacticType(s.type)).collect(Collectors.toList());
			return new TupleType(types);
//			RecordType rT=(RecordType)type;
//			ValueSymbol syms[]=rT.symbols;
//			List<Parameter> params=Arrays.stream(syms).map(s -> new Parameter(new Identifier(s.ident.string), 
//					canonic2syntacticType(s.type))).collect(Collectors.toList());
//			return new RecordType(params);
		}
		if(type instanceof SetType)
		{
			SetType sT=(SetType)type;
			Type type1=canonic2syntacticType(sT.type);
			return new SetType(type1);
		}
		if(type instanceof SubType)
		{
			SubType sT=(SubType)type;
			return new IdentifierType(new Identifier(sT.ident.string));
		}
		if(type instanceof RecursiveType)
		{
			RecursiveType rT=(RecursiveType)type;
			return new IdentifierType(new Identifier(rT.ident.string));
		}			
		assert(false);
		return null;
	}
	
	/***************************************************************************
	 * Computes nested forall expressions.
	 * @param qv The variables that shall be used for quantification.
	 * @param exp The expression which shall be quantified
	 * @return Nested Forall expressions where each forall introduces only one variable.
	 * 	The first forall introduces the first variable from qv the second the second,...
	 **************************************************************************/
	static Expression constructNestedForallExpression(QuantifiedVariable qv,Expression exp)
	{
		for(int i=qv.qvcore.length-1;i>=0;i--)
		{
			List<QuantifiedVariableCore> corelst=new ArrayList<AST.QuantifiedVariableCore>(1);
			corelst.add(qv.qvcore[i]);
			QuantifiedVariable var=new QuantifiedVariable(corelst, null);
			exp=new ForallExp(var, exp);
		}
		return exp;
	}
	
	/***************************************************************************
	 * Computes nested exists expressions.
	 * @param qv The variables that shall be used for quantification.
	 * @param exp The expression which shall be quantified
	 * @return Nested Exists expressions where each exists introduces only one variable.
	 * 	The first exists introduces the first variable from qv the second the second,...
	 **************************************************************************/
	static Expression constructNestedExistsExpression(QuantifiedVariable qv,Expression exp)
	{
		for(int i=qv.qvcore.length-1;i>=0;i--)
		{
			List<QuantifiedVariableCore> corelst=new ArrayList<AST.QuantifiedVariableCore>(1);
			corelst.add(qv.qvcore[i]);
			QuantifiedVariable var=new QuantifiedVariable(corelst, null);
			exp=new ExistsExp(var, exp);
		}
		return exp;
	}
	
	/***************************************************************************
	 * Computes the value symbols of all free variables in exp.
	 * @param exp The expression whose free variables we want to compute
	 * @param values A set of valueSymbols which are the symbols of non variables.
	 * @return The ValueSymbol of all identifiers which are not bound by quantifiers or lets
	 * 		from the expression exp minus the symbols from values.
	 **************************************************************************/
	static List<ValueSymbol> getTypedFreeVariablesIgnoreValues(Expression exp,Set<ValueSymbol> values)
	{
		Set<ValueSymbol> free=TypedFreeVariables.compute(exp);
		free.removeAll(values);
		return sortSymbols(free);
	}
	
	/***************************************************************************
	 * Computes the value symbols of all free variables in exp.
	 * Consider the symbols given by additionalSymbols as free symbols
	 * @param exp The expression whose free variables we want to compute
	 * @param values A set of valueSymbols which are the symbols of non variables.
	 * @param additionalSymbols valueSymbols which shall be considered as free variables
	 * @return The ValueSymbol of all identifiers which are not bound by quantifiers or lets
	 * 		from the expression exp plus the symbols from additionalSymbols minus the symbols from values
	 **************************************************************************/
	static List<ValueSymbol> getTypedFreeVariablesIgnoreValues(Expression exp,Set<ValueSymbol> values,Collection<ValueSymbol> additionalSymbols)
	{
		Set<ValueSymbol> free=TypedFreeVariables.compute(exp);
		free.addAll(additionalSymbols);
		free.removeAll(values);
		return sortSymbols(free);
	}
	
	/***************************************************************************
	 * Sorts a set of variables with respect to the option SMT.qvOrdering
	 * @param syms A collection of value symbols that shall be sorted
	 * @return A list of value symbols containing the elements of syms
	 **************************************************************************/
	static List<ValueSymbol> sortSymbols(Collection<ValueSymbol> syms)
	{
		List<ValueSymbol> freelst=new ArrayList<ValueSymbol>(syms);
		switch(SMT.qvOrdering)
  	{
  	case NoOrdering:
  		break;
  	case OrderAscending_TypeSize:
  		freelst.sort(new ValueSymAscendingComparator());
  		break;
  	case OrderDescending_TypeSize:
  		freelst.sort(new ValueSymDescendingComparator());
  		break;
  	}
		return freelst;
	}
	
	/***************************************************************************
	 * Compares QuantifiedVariablesCore with respect to the associated
	 * value symbols
	 **************************************************************************/
	static class QVAscendingComparator implements Comparator<QuantifiedVariableCore>
  {
		private static ValueSymAscendingComparator symComp=new ValueSymAscendingComparator();		
		
		@Override
		public int compare(QuantifiedVariableCore o1, QuantifiedVariableCore o2) 
		{
			return symComp.compare((ValueSymbol)o1.ident.getSymbol(), (ValueSymbol)o2.ident.getSymbol());
		} 	
  }
  
	/***************************************************************************
	 * Compares ValueSymbol with respect to the size of the type
	 * of the associated value symbol.
	 * If the type sizes are equal the names of the identifiers
	 * of the symbol are compared
	 **************************************************************************/
	static class ValueSymAscendingComparator implements Comparator<ValueSymbol>
  {

		@Override
		public int compare(ValueSymbol o1, ValueSymbol o2) 
		{
			Type ty1=o1.type;
			Type ty2=o2.type;
			int comp=Long.compare(ty1.getSize(), ty2.getSize());
			if(comp==0)
				return o1.ident.string.compareTo(o2.ident.string);
			return comp;
		} 	
  }
  
	/***************************************************************************
	 * Compares QuantifiedVariablesCore with respect to the associated
	 * value symbols
	 **************************************************************************/
	static class QVDescendingComparator implements Comparator<QuantifiedVariableCore>
  {
		private static ValueSymDescendingComparator symComp=new ValueSymDescendingComparator();		
		@Override
		public int compare(QuantifiedVariableCore o1, QuantifiedVariableCore o2) 
		{
			return symComp.compare((ValueSymbol)o1.ident.getSymbol(), (ValueSymbol)o2.ident.getSymbol());
		} 	
  }
  
	/***************************************************************************
	 * Compares ValueSymbol with respect to the size of the type
	 * of the associated value symbol.
	 * If the type sizes are equal the names of the identifiers
	 * of the symbol are compared
	 **************************************************************************/
	static class ValueSymDescendingComparator implements Comparator<ValueSymbol>
  {

		@Override
		public int compare(ValueSymbol o1, ValueSymbol o2) 
		{
			Type ty1=o1.type;
			Type ty2=o2.type;
			int comp=-Long.compare(ty1.getSize(), ty2.getSize());
			if(comp==0)
				return o1.ident.string.compareTo(o2.ident.string);
			return comp;
		} 	
  }
	
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
